# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the :class:`iris.coord_systems.RotatedPole` class."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock

import cartopy
import cartopy.crs as ccrs
from iris.coord_systems import RotatedGeogCS


class Test_init(tests.IrisTest):
    def setUp(self):
        self.pole_lon = 171.77
        self.pole_lat = 49.55
        self.rotation_about_new_pole = 180.0
        self.rp_crs = RotatedGeogCS(
            self.pole_lat, self.pole_lon, self.rotation_about_new_pole
        )

    def test_crs_creation(self):
        self.assertEqual(self.pole_lon, self.rp_crs.grid_north_pole_longitude)
        self.assertEqual(self.pole_lat, self.rp_crs.grid_north_pole_latitude)
        self.assertEqual(
            self.rotation_about_new_pole, self.rp_crs.north_pole_grid_longitude
        )

    def test_as_cartopy_crs(self):
        if cartopy.__version__ < "0.12":
            with mock.patch("warnings.warn") as warn:
                accrs = self.rp_crs.as_cartopy_crs()
                self.assertEqual(warn.call_count, 1)
        else:
            accrs = self.rp_crs.as_cartopy_crs()
            expected = ccrs.RotatedGeodetic(
                self.pole_lon, self.pole_lat, self.rotation_about_new_pole
            )
            self.assertEqual(
                sorted(accrs.proj4_init.split(" +")),
                sorted(expected.proj4_init.split(" +")),
            )

    def test_as_cartopy_projection(self):
        if cartopy.__version__ < "0.12":
            with mock.patch("warnings.warn") as warn:
                _ = self.rp_crs.as_cartopy_projection()
                self.assertEqual(warn.call_count, 1)
        else:
            accrsp = self.rp_crs.as_cartopy_projection()
            expected = ccrs.RotatedPole(
                self.pole_lon, self.pole_lat, self.rotation_about_new_pole
            )
            self.assertEqual(
                sorted(accrsp.proj4_init.split(" +")),
                sorted(expected.proj4_init.split(" +")),
            )

    def _check_crs_default(self, crs):
        # Check for property defaults when no kwargs options are set.
        # NOTE: except ellipsoid, which is done elsewhere.
        self.assertEqualAndKind(crs.north_pole_grid_longitude, 0.0)

    def test_optional_args_missing(self):
        # Check that unused 'north_pole_grid_longitude' defaults to 0.0.
        crs = RotatedGeogCS(self.pole_lon, self.pole_lat)
        self._check_crs_default(crs)

    def test_optional_args_None(self):
        # Check that 'north_pole_grid_longitude=None' defaults to 0.0.
        crs = RotatedGeogCS(
            self.pole_lon, self.pole_lat, north_pole_grid_longitude=None
        )
        self._check_crs_default(crs)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test iris.util

"""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

import inspect

import cf_units
import numpy as np

from io import StringIO

import iris.analysis
import iris.coords
import iris.tests.stock as stock
import iris.util


class TestMonotonic(tests.IrisTest):
    def assertMonotonic(self, array, direction=None, **kwargs):
        if direction is not None:
            mono, dir = iris.util.monotonic(
                array, return_direction=True, **kwargs
            )
            if not mono:
                self.fail("Array was not monotonic:/n %r" % array)
            if dir != np.sign(direction):
                self.fail(
                    "Array was monotonic but not in the direction expected:"
                    "/n  + requested direction: %s/n  + resultant direction: %s"
                    % (direction, dir)
                )
        else:
            mono = iris.util.monotonic(array, **kwargs)
            if not mono:
                self.fail("Array was not monotonic:/n %r" % array)

    def assertNotMonotonic(self, array, **kwargs):
        mono = iris.util.monotonic(array, **kwargs)
        if mono:
            self.fail("Array was monotonic when it shouldn't be:/n %r" % array)

    def test_monotonic_pve(self):
        a = np.array([3, 4, 5.3])
        self.assertMonotonic(a)
        self.assertMonotonic(a, direction=1)

        # test the reverse for negative monotonic.
        a = a[::-1]
        self.assertMonotonic(a)
        self.assertMonotonic(a, direction=-1)

    def test_not_monotonic(self):
        b = np.array([3, 5.3, 4])
        self.assertNotMonotonic(b)

    def test_monotonic_strict(self):
        b = np.array([3, 5.3, 4])
        self.assertNotMonotonic(b, strict=True)
        self.assertNotMonotonic(b)

        b = np.array([3, 5.3, 5.3])
        self.assertNotMonotonic(b, strict=True)
        self.assertMonotonic(b, direction=1)

        b = b[::-1]
        self.assertNotMonotonic(b, strict=True)
        self.assertMonotonic(b, direction=-1)

        b = np.array([0.0])
        self.assertRaises(ValueError, iris.util.monotonic, b)
        self.assertRaises(ValueError, iris.util.monotonic, b, strict=True)

        b = np.array([0.0, 0.0])
        self.assertNotMonotonic(b, strict=True)
        self.assertMonotonic(b)


class TestClipString(tests.IrisTest):
    def setUp(self):
        self.test_string = "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum."
        self.rider = "**^^**$$..--__"  # A good chance at being unique and not in the string to be tested!

    def test_oversize_string(self):
        # Test with a clip length that means the string will be clipped

        clip_length = 109
        result = iris.util.clip_string(
            self.test_string, clip_length, self.rider
        )

        # Check the length is between what we requested ( + rider length) and the length of the original string
        self.assertTrue(
            clip_length + len(self.rider)
            <= len(result)
            < len(self.test_string),
            "String was not clipped.",
        )

        # Also test the rider was added
        self.assertTrue(
            self.rider in result,
            "Rider was not added to the string when it should have been.",
        )

    def test_undersize_string(self):
        # Test with a clip length that is longer than the string

        clip_length = 10999
        result = iris.util.clip_string(
            self.test_string, clip_length, self.rider
        )
        self.assertEqual(
            len(result),
            len(self.test_string),
            "String was clipped when it should not have been.",
        )

        # Also test that no rider was added on the end if the string was not clipped
        self.assertFalse(
            self.rider in result,
            "Rider was adding to the string when it should not have been.",
        )

    def test_invalid_clip_lengths(self):
        # Clip values less than or equal to zero are not valid
        for clip_length in [0, -100]:
            result = iris.util.clip_string(
                self.test_string, clip_length, self.rider
            )
            self.assertEqual(
                len(result),
                len(self.test_string),
                "String was clipped when it should not have been.",
            )

    def test_default_values(self):
        # Get the default values specified in the function
        argspec = inspect.getargspec(iris.util.clip_string)
        arg_dict = dict(zip(argspec.args[-2:], argspec.defaults))

        result = iris.util.clip_string(
            self.test_string, arg_dict["clip_length"], arg_dict["rider"]
        )

        self.assertLess(
            len(result), len(self.test_string), "String was not clipped."
        )

        rider_returned = result[-len(arg_dict["rider"]) :]
        self.assertEqual(
            rider_returned, arg_dict["rider"], "Default rider was not applied."
        )

    def test_trim_string_with_no_spaces(self):

        clip_length = 200
        no_space_string = "a" * 500

        # Since this string has no spaces, clip_string will not be able to gracefully clip it
        # but will instead clip it exactly where the user specified
        result = iris.util.clip_string(
            no_space_string, clip_length, self.rider
        )

        expected_length = clip_length + len(self.rider)

        # Check the length of the returned string is equal to clip length + length of rider
        self.assertEqual(
            len(result),
            expected_length,
            "Mismatch in expected length of clipped string. Length was %s, "
            "expected value is %s" % (len(result), expected_length),
        )


@tests.skip_data
class TestDescribeDiff(iris.tests.IrisTest):
    def test_identical(self):
        test_cube_a = stock.realistic_4d()
        test_cube_b = stock.realistic_4d()

        return_sio = StringIO()
        iris.util.describe_diff(
            test_cube_a, test_cube_b, output_file=return_sio
        )
        return_str = return_sio.getvalue()

        self.assertString(return_str, "compatible_cubes.str.txt")

    def test_different(self):
        # test incompatible attributes
        test_cube_a = stock.realistic_4d()
        test_cube_b = stock.realistic_4d()

        test_cube_a.attributes["Conventions"] = "CF-1.5"
        test_cube_b.attributes["Conventions"] = "CF-1.6"

        return_sio = StringIO()
        iris.util.describe_diff(
            test_cube_a, test_cube_b, output_file=return_sio
        )
        return_str = return_sio.getvalue()

        self.assertString(return_str, "incompatible_attr.str.txt")

        # test incompatible names
        test_cube_a = stock.realistic_4d()
        test_cube_b = stock.realistic_4d()

        test_cube_a.standard_name = "relative_humidity"

        return_sio = StringIO()
        iris.util.describe_diff(
            test_cube_a, test_cube_b, output_file=return_sio
        )
        return_str = return_sio.getvalue()

        self.assertString(return_str, "incompatible_name.str.txt")

        # test incompatible unit
        test_cube_a = stock.realistic_4d()
        test_cube_b = stock.realistic_4d()

        test_cube_a.units = cf_units.Unit("m")

        return_sio = StringIO()
        iris.util.describe_diff(
            test_cube_a, test_cube_b, output_file=return_sio
        )
        return_str = return_sio.getvalue()

        self.assertString(return_str, "incompatible_unit.str.txt")

        # test incompatible methods
        test_cube_a = stock.realistic_4d()
        test_cube_b = stock.realistic_4d().collapsed(
            "model_level_number", iris.analysis.MEAN
        )

        return_sio = StringIO()
        iris.util.describe_diff(
            test_cube_a, test_cube_b, output_file=return_sio
        )
        return_str = return_sio.getvalue()

        self.assertString(return_str, "incompatible_meth.str.txt")

    def test_output_file(self):
        # test incompatible attributes
        test_cube_a = stock.realistic_4d()
        test_cube_b = stock.realistic_4d().collapsed(
            "model_level_number", iris.analysis.MEAN
        )

        test_cube_a.attributes["Conventions"] = "CF-1.5"
        test_cube_b.attributes["Conventions"] = "CF-1.6"
        test_cube_a.standard_name = "relative_humidity"
        test_cube_a.units = cf_units.Unit("m")

        with self.temp_filename() as filename:
            with open(filename, "w") as f:
                iris.util.describe_diff(
                    test_cube_a, test_cube_b, output_file=f
                )
                f.close()

            self.assertFilesEqual(filename, "incompatible_cubes.str.txt")


@tests.skip_data
class TestAsCompatibleShape(tests.IrisTest):
    def test_slice(self):
        cube = tests.stock.realistic_4d()
        sliced = cube[1, :, 2, :-2]
        expected = cube[1:2, :, 2:3, :-2]
        res = iris.util.as_compatible_shape(sliced, cube)
        self.assertEqual(res, expected)

    def test_transpose(self):
        cube = tests.stock.realistic_4d()
        transposed = cube.copy()
        transposed.transpose()
        expected = cube
        res = iris.util.as_compatible_shape(transposed, cube)
        self.assertEqual(res, expected)

    def test_slice_and_transpose(self):
        cube = tests.stock.realistic_4d()
        sliced_and_transposed = cube[1, :, 2, :-2]
        sliced_and_transposed.transpose()
        expected = cube[1:2, :, 2:3, :-2]
        res = iris.util.as_compatible_shape(sliced_and_transposed, cube)
        self.assertEqual(res, expected)

    def test_collapsed(self):
        cube = tests.stock.realistic_4d()
        collapsed = cube.collapsed("model_level_number", iris.analysis.MEAN)
        expected_shape = list(cube.shape)
        expected_shape[1] = 1
        expected_data = collapsed.data.reshape(expected_shape)
        res = iris.util.as_compatible_shape(collapsed, cube)
        self.assertCML(
            res, ("util", "as_compatible_shape_collapsed.cml"), checksum=False
        )
        self.assertMaskedArrayEqual(expected_data, res.data)

    def test_reduce_dimensionality(self):
        # Test that as_compatible_shape() can demote
        # length one dimensions to scalars.
        cube = tests.stock.realistic_4d()
        src = cube[:, 2:3]
        expected = reduced = cube[:, 2]
        res = iris.util.as_compatible_shape(src, reduced)
        self.assertEqual(res, expected)

    def test_anonymous_dims(self):
        cube = tests.stock.realistic_4d()
        # Move all coords from dim_coords to aux_coords.
        for coord in cube.dim_coords:
            dim = cube.coord_dims(coord)
            cube.remove_coord(coord)
            cube.add_aux_coord(coord, dim)

        sliced = cube[1, :, 2, :-2]
        expected = cube[1:2, :, 2:3, :-2]
        res = iris.util.as_compatible_shape(sliced, cube)
        self.assertEqual(res, expected)

    def test_scalar_auxcoord(self):
        def dim_to_aux(cube, coord_name):
            """Convert coordinate on cube from DimCoord to AuxCoord."""
            coord = cube.coord(coord_name)
            coord = iris.coords.AuxCoord.from_coord(coord)
            cube.replace_coord(coord)

        cube = tests.stock.realistic_4d()
        src = cube[:, :, 3]
        dim_to_aux(src, "grid_latitude")
        expected = cube[:, :, 3:4]
        dim_to_aux(expected, "grid_latitude")
        res = iris.util.as_compatible_shape(src, cube)
        self.assertEqual(res, expected)

    def test_2d_auxcoord_transpose(self):
        dim_coord1 = iris.coords.DimCoord(range(3), long_name="first_dim")
        dim_coord2 = iris.coords.DimCoord(range(4), long_name="second_dim")
        aux_coord_2d = iris.coords.AuxCoord(
            np.arange(12).reshape(3, 4), long_name="spanning"
        )
        aux_coord_2d_T = iris.coords.AuxCoord(
            np.arange(12).reshape(3, 4).T, long_name="spanning"
        )
        src = iris.cube.Cube(
            np.ones((3, 4)),
            dim_coords_and_dims=[(dim_coord1, 0), (dim_coord2, 1)],
            aux_coords_and_dims=[(aux_coord_2d, (0, 1))],
        )
        target = iris.cube.Cube(
            np.ones((4, 3)),
            dim_coords_and_dims=[(dim_coord1, 1), (dim_coord2, 0)],
            aux_coords_and_dims=[(aux_coord_2d_T, (0, 1))],
        )

        res = iris.util.as_compatible_shape(src, target)
        self.assertEqual(res[0], target[0])


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Defines a Trajectory class, and a routine to extract a sub-cube along a
trajectory.

"""

import math

import numpy as np
from scipy.spatial import cKDTree

import iris.analysis
import iris.coord_systems
import iris.coords

from iris.analysis import Linear
from iris.analysis._interpolation import snapshot_grid
from iris.util import _meshgrid


class _Segment:
    """A single trajectory line segment: Two points, as described in the
    Trajectory class."""

    def __init__(self, p0, p1):
        # check keys
        if sorted(p0.keys()) != sorted(p1.keys()):
            raise ValueError("keys do not match")

        self.pts = [p0, p1]

        # calculate our length
        squares = 0
        for key in self.pts[0].keys():
            delta = self.pts[1][key] - self.pts[0][key]
            squares += delta * delta
        self.length = math.sqrt(squares)


class Trajectory:
    """A series of given waypoints with pre-calculated sample points."""

    def __init__(self, waypoints, sample_count=10):
        """
        Defines a trajectory using a sequence of waypoints.

        For example::

            waypoints = [{'latitude': 45, 'longitude': -60},
                         {'latitude': 45, 'longitude': 0}]
            Trajectory(waypoints)

        .. note:: All the waypoint dictionaries must contain the same
            coordinate names.

        Args:

        * waypoints
            A sequence of dictionaries, mapping coordinate names to values.

        Kwargs:

        * sample_count
            The number of sample positions to use along the trajectory.

        """
        self.waypoints = waypoints
        self.sample_count = sample_count

        # create line segments from the waypoints
        segments = [
            _Segment(self.waypoints[i], self.waypoints[i + 1])
            for i in range(len(self.waypoints) - 1)
        ]

        # calculate our total length
        self.length = sum([seg.length for seg in segments])

        # generate our sampled points
        #: The trajectory points, as dictionaries of {coord_name: value}.
        self.sampled_points = []
        sample_step = self.length / (self.sample_count - 1)

        # start with the first segment
        cur_seg_i = 0
        cur_seg = segments[cur_seg_i]
        len_accum = cur_seg.length
        for p in range(self.sample_count):

            # calculate the sample position along our total length
            sample_at_len = p * sample_step

            # skip forward to the containing segment
            while len_accum < sample_at_len and cur_seg_i < len(segments):
                cur_seg_i += 1
                cur_seg = segments[cur_seg_i]
                len_accum += cur_seg.length

            # how far through the segment is our sample point?
            seg_start_len = len_accum - cur_seg.length
            seg_frac = (sample_at_len - seg_start_len) / cur_seg.length

            # sample each coordinate in this segment, to create a new
            # sampled point
            new_sampled_point = {}
            for key in cur_seg.pts[0].keys():
                seg_coord_delta = cur_seg.pts[1][key] - cur_seg.pts[0][key]
                new_sampled_point.update(
                    {key: cur_seg.pts[0][key] + seg_frac * seg_coord_delta}
                )

            # add this new sampled point
            self.sampled_points.append(new_sampled_point)

    def __repr__(self):
        return "Trajectory(%s, sample_count=%s)" % (
            self.waypoints,
            self.sample_count,
        )

    def _get_interp_points(self):
        """
        Translate `self.sampled_points` to the format expected by the
        interpolator.

        Returns:
            `self.sampled points` in the format required by
            `:func:`~iris.analysis.trajectory.interpolate`.

        """
        points = {
            k: [point_dict[k] for point_dict in self.sampled_points]
            for k in self.sampled_points[0].keys()
        }
        return [(k, v) for k, v in points.items()]

    def _src_cube_anon_dims(self, cube):
        """
        A helper method to locate the index of anonymous dimensions on the
        interpolation target, ``cube``.

        Returns:
            The index of any anonymous dimensions in ``cube``.

        """
        named_dims = [cube.coord_dims(c)[0] for c in cube.dim_coords]
        return list(set(range(cube.ndim)) - set(named_dims))

    def interpolate(self, cube, method=None):
        """
        Calls :func:`~iris.analysis.trajectory.interpolate` to interpolate
        ``cube`` on the defined trajectory.

        Assumes that the coordinate names supplied in the waypoints
        dictionaries match to coordinate names in `cube`, and that points are
        supplied in the same coord_system as in `cube`, where appropriate (i.e.
        for horizontal coordinate points).

        Args:

        * cube
             The source Cube to interpolate.

        Kwargs:

        * method:
            The interpolation method to use; "linear" (default) or "nearest".
            Only nearest is available when specifying multi-dimensional
            coordinates.

        """
        sample_points = self._get_interp_points()
        interpolated_cube = interpolate(cube, sample_points, method=method)
        # Add an "index" coord to name the anonymous dimension produced by
        # the interpolation, if present.
        if len(interpolated_cube.dim_coords) < interpolated_cube.ndim:
            # Add a new coord `index` to describe the new dimension created by
            # interpolating.
            index_coord = iris.coords.DimCoord(
                range(self.sample_count), long_name="index"
            )
            # Make sure anonymous dims in `cube` do not mistakenly get labelled
            # as the new `index` dimension created by interpolating.
            src_anon_dims = self._src_cube_anon_dims(cube)
            interp_anon_dims = self._src_cube_anon_dims(interpolated_cube)
            (anon_dim_index,) = list(
                set(interp_anon_dims) - set(src_anon_dims)
            )
            # Add the new coord to the interpolated cube.
            interpolated_cube.add_dim_coord(index_coord, anon_dim_index)
        return interpolated_cube


def interpolate(cube, sample_points, method=None):
    """
    Extract a sub-cube at the given n-dimensional points.

    Args:

    * cube
        The source Cube.

    * sample_points
        A sequence of coordinate (name) - values pairs.

    Kwargs:

    * method
        Request "linear" interpolation (default) or "nearest" neighbour.
        Only nearest neighbour is available when specifying multi-dimensional
        coordinates.


    For example::

        sample_points = [('latitude', [45, 45, 45]),
        ('longitude', [-60, -50, -40])]
        interpolated_cube = interpolate(cube, sample_points)

    """
    if method not in [None, "linear", "nearest"]:
        raise ValueError("Unhandled interpolation specified : %s" % method)

    # Convert any coordinate names to coords
    points = []
    for coord, values in sample_points:
        if isinstance(coord, str):
            coord = cube.coord(coord)
        points.append((coord, values))
    sample_points = points

    # Do all value sequences have the same number of values?
    coord, values = sample_points[0]
    trajectory_size = len(values)
    for coord, values in sample_points[1:]:
        if len(values) != trajectory_size:
            raise ValueError("Lengths of coordinate values are inconsistent.")

    # Which dimensions are we squishing into the last dimension?
    squish_my_dims = set()
    for coord, values in sample_points:
        dims = cube.coord_dims(coord)
        for dim in dims:
            squish_my_dims.add(dim)

    # Derive the new cube's shape by filtering out all the dimensions we're
    # about to sample,
    # and then adding a new dimension to accommodate all the sample points.
    remaining = [
        (dim, size)
        for dim, size in enumerate(cube.shape)
        if dim not in squish_my_dims
    ]
    new_data_shape = [size for dim, size in remaining]
    new_data_shape.append(trajectory_size)

    # Start with empty data and then fill in the "column" of values for each
    # trajectory point.
    new_cube = iris.cube.Cube(np.empty(new_data_shape))
    new_cube.metadata = cube.metadata

    # Derive the mapping from the non-trajectory source dimensions to their
    # corresponding destination dimensions.
    remaining_dims = [dim for dim, size in remaining]
    dimension_remap = {dim: i for i, dim in enumerate(remaining_dims)}

    # Record a mapping from old coordinate IDs to new coordinates,
    # for subsequent use in creating updated aux_factories.
    coord_mapping = {}

    # Create all the non-squished coords
    for coord in cube.dim_coords:
        src_dims = cube.coord_dims(coord)
        if squish_my_dims.isdisjoint(src_dims):
            dest_dims = [dimension_remap[dim] for dim in src_dims]
            new_coord = coord.copy()
            new_cube.add_dim_coord(new_coord, dest_dims)
            coord_mapping[id(coord)] = new_coord
    for coord in cube.aux_coords:
        src_dims = cube.coord_dims(coord)
        if squish_my_dims.isdisjoint(src_dims):
            dest_dims = [dimension_remap[dim] for dim in src_dims]
            new_coord = coord.copy()
            new_cube.add_aux_coord(new_coord, dest_dims)
            coord_mapping[id(coord)] = new_coord

    # Create all the squished (non derived) coords, not filled in yet.
    trajectory_dim = len(remaining_dims)
    for coord in cube.dim_coords + cube.aux_coords:
        src_dims = cube.coord_dims(coord)
        if not squish_my_dims.isdisjoint(src_dims):
            points = np.array([coord.points.flatten()[0]] * trajectory_size)
            new_coord = iris.coords.AuxCoord(
                points,
                var_name=coord.var_name,
                standard_name=coord.standard_name,
                long_name=coord.long_name,
                units=coord.units,
                bounds=None,
                attributes=coord.attributes,
                coord_system=coord.coord_system,
            )
            new_cube.add_aux_coord(new_coord, trajectory_dim)
            coord_mapping[id(coord)] = new_coord

    for factory in cube.aux_factories:
        new_cube.add_aux_factory(factory.updated(coord_mapping))

    # Are the given coords all 1-dimensional? (can we do linear interp?)
    for coord, values in sample_points:
        if coord.ndim > 1:
            if method == "linear":
                msg = (
                    "Cannot currently perform linear interpolation for "
                    "multi-dimensional coordinates."
                )
                raise iris.exceptions.CoordinateMultiDimError(msg)
            method = "nearest"
            break

    if method in ["linear", None]:
        for i in range(trajectory_size):
            point = [(coord, values[i]) for coord, values in sample_points]
            column = cube.interpolate(point, Linear())
            new_cube.data[..., i] = column.data
            # Fill in the empty squashed (non derived) coords.
            for column_coord in column.dim_coords + column.aux_coords:
                src_dims = cube.coord_dims(column_coord)
                if not squish_my_dims.isdisjoint(src_dims):
                    if len(column_coord.points) != 1:
                        msg = "Expected to find exactly one point. Found {}."
                        raise Exception(msg.format(column_coord.points))
                    new_cube.coord(column_coord.name()).points[
                        i
                    ] = column_coord.points[0]

    elif method == "nearest":
        # Use a cache with _nearest_neighbour_indices_ndcoords()
        cache = {}
        column_indexes = _nearest_neighbour_indices_ndcoords(
            cube, sample_points, cache=cache
        )

        # Construct "fancy" indexes, so we can create the result data array in
        # a single numpy indexing operation.
        # ALSO: capture the index range in each dimension, so that we can fetch
        # only a required (square) sub-region of the source data.
        fancy_source_indices = []
        region_slices = []
        n_index_length = len(column_indexes[0])
        dims_reduced = [False] * n_index_length
        for i_ind in range(n_index_length):
            contents = [column_index[i_ind] for column_index in column_indexes]
            each_used = [content != slice(None) for content in contents]
            if np.all(each_used):
                # This dimension is addressed : use a list of indices.
                dims_reduced[i_ind] = True
                # Select the region by min+max indices.
                start_ind = np.min(contents)
                stop_ind = 1 + np.max(contents)
                region_slice = slice(start_ind, stop_ind)
                # Record point indices with start subtracted from all of them.
                fancy_index = list(np.array(contents) - start_ind)
            elif not np.any(each_used):
                # This dimension is not addressed by the operation.
                # Use a ":" as the index.
                fancy_index = slice(None)
                # No sub-region selection for this dimension.
                region_slice = slice(None)
            else:
                # Should really never happen, if _ndcoords is right.
                msg = (
                    "Internal error in trajectory interpolation : point "
                    "selection indices should all have the same form."
                )
                raise ValueError(msg)

            fancy_source_indices.append(fancy_index)
            region_slices.append(region_slice)

        # Fetch the required (square-section) region of the source data.
        # NOTE: This is not quite as good as only fetching the individual
        # points used, but it avoids creating a sub-cube for each point,
        # which is very slow, especially when points are re-used a lot ...
        source_area_indices = tuple(region_slices)
        source_data = cube[source_area_indices].data

        # Transpose source data before indexing it to get the final result.
        # Because.. the fancy indexing will replace the indexed (horizontal)
        # dimensions with a new single dimension over trajectory points.
        # Move those dimensions to the end *first* : this ensures that the new
        # dimension also appears at the end, which is where we want it.
        # Make a list of dims with the reduced ones last.
        dims_reduced = np.array(dims_reduced)
        dims_order = np.arange(n_index_length)
        dims_order = np.concatenate(
            (dims_order[~dims_reduced], dims_order[dims_reduced])
        )
        # Rearrange the data dimensions and the fancy indices into that order.
        source_data = source_data.transpose(dims_order)
        fancy_source_indices = [
            fancy_source_indices[i_dim] for i_dim in dims_order
        ]

        # Apply the fancy indexing to get all the result data points.
        source_data = source_data[tuple(fancy_source_indices)]

        # "Fix" problems with missing datapoints producing odd values
        # when copied from a masked into an unmasked array.
        # TODO: proper masked data handling.
        if np.ma.isMaskedArray(source_data):
            # This is **not** proper mask handling, because we cannot produce a
            # masked result, but it ensures we use a "filled" version of the
            # input in this case.
            source_data = source_data.filled()
        new_cube.data[:] = source_data
        # NOTE: we assign to "new_cube.data[:]" and *not* just "new_cube.data",
        # because the existing code produces a default dtype from 'np.empty'
        # instead of preserving the input dtype.
        # TODO: maybe this should be fixed -- i.e. to preserve input dtype ??

        # Fill in the empty squashed (non derived) coords.
        column_coords = [
            coord
            for coord in cube.dim_coords + cube.aux_coords
            if not squish_my_dims.isdisjoint(cube.coord_dims(coord))
        ]
        new_cube_coords = [
            new_cube.coord(column_coord.name())
            for column_coord in column_coords
        ]
        all_point_indices = np.array(column_indexes)
        single_point_test_cube = cube[column_indexes[0]]
        for new_cube_coord, src_coord in zip(new_cube_coords, column_coords):
            # Check structure of the indexed coord (at one selected point).
            point_coord = single_point_test_cube.coord(src_coord)
            if len(point_coord.points) != 1:
                msg = (
                    "Coord {} at one x-y position has the shape {}, "
                    "instead of being a single point. "
                )
                raise ValueError(msg.format(src_coord.name(), src_coord.shape))

            # Work out which indices apply to the input coord.
            # NOTE: we know how to index the source cube to get a cube with a
            # single point for each coord, but this is very inefficient.
            # So here, we translate cube indexes into *coord* indexes.
            src_coord_dims = cube.coord_dims(src_coord)
            fancy_coord_index_arrays = [
                list(all_point_indices[:, src_dim])
                for src_dim in src_coord_dims
            ]

            # Fill the new coord with all the correct points from the old one.
            new_cube_coord.points = src_coord.points[
                tuple(fancy_coord_index_arrays)
            ]
            # NOTE: the new coords do *not* have bounds.

    return new_cube


def _ll_to_cart(lon, lat):
    # Based on cartopy.img_transform.ll_to_cart().
    x = np.sin(np.deg2rad(90 - lat)) * np.cos(np.deg2rad(lon))
    y = np.sin(np.deg2rad(90 - lat)) * np.sin(np.deg2rad(lon))
    z = np.cos(np.deg2rad(90 - lat))
    return (x, y, z)


def _cartesian_sample_points(sample_points, sample_point_coord_names):
    """
    Replace geographic lat/lon with cartesian xyz.
    Generates coords suitable for nearest point calculations with
    `scipy.spatial.cKDTree`.

    Args:

    * sample_points[coord][datum]:
        list of sample_positions for each datum, formatted for fast use of
        :func:`_ll_to_cart()`.

    * sample_point_coord_names[coord]:
        list of n coord names

    Returns:
        list of [x,y,z,t,etc] positions, formatted for kdtree.

    """
    # Find lat and lon coord indices
    i_lat = i_lon = None
    i_non_latlon = list(range(len(sample_point_coord_names)))
    for i, name in enumerate(sample_point_coord_names):
        if "latitude" in name:
            i_lat = i
            i_non_latlon.remove(i_lat)
        if "longitude" in name:
            i_lon = i
            i_non_latlon.remove(i_lon)

    if i_lat is None or i_lon is None:
        return sample_points.transpose()

    num_points = len(sample_points[0])
    cartesian_points = [None] * num_points

    # Get the point coordinates without the latlon
    for p in range(num_points):
        cartesian_points[p] = [sample_points[c][p] for c in i_non_latlon]

    # Add cartesian xyz coordinates from latlon
    x, y, z = _ll_to_cart(sample_points[i_lon], sample_points[i_lat])
    for p in range(num_points):
        cartesian_point = cartesian_points[p]
        cartesian_point.append(x[p])
        cartesian_point.append(y[p])
        cartesian_point.append(z[p])

    return cartesian_points


def _nearest_neighbour_indices_ndcoords(cube, sample_points, cache=None):
    """
    Returns the indices to select the data value(s) closest to the given
    coordinate point values.

    'sample_points' is of the form [[coord-or-coord-name, point-value(s)]*].
    The lengths of all the point-values sequences must be equal.

    This function is adapted for points sampling a multi-dimensional coord,
    and can currently only do nearest neighbour interpolation.

    Because this function can be slow for multidimensional coordinates,
    a 'cache' dictionary can be provided by the calling code.

    .. Note::

        If the points are longitudes/latitudes, these are handled correctly as
        points on the sphere, but the values must be in 'degrees'.

    Developer notes:
    A "sample space cube" is made which only has the coords and dims we are
    sampling on.
    We get the nearest neighbour using this sample space cube.

    """
    if sample_points:
        try:
            coord, value = sample_points[0]
        except (KeyError, ValueError):
            emsg = (
                "Sample points must be a list of "
                "(coordinate, value) pairs, got {!r}."
            )
            raise TypeError(emsg.format(sample_points))

    # Convert names to coords in sample_point and reformat sample point values
    # for use in `_cartesian_sample_points()`.
    coord_values = []
    sample_point_coords = []
    sample_point_coord_names = []
    ok_coord_ids = set(map(id, cube.dim_coords + cube.aux_coords))
    for coord, value in sample_points:
        coord = cube.coord(coord)
        if id(coord) not in ok_coord_ids:
            msg = (
                "Invalid sample coordinate {!r}: derived coordinates are"
                " not allowed.".format(coord.name())
            )
            raise ValueError(msg)
        sample_point_coords.append(coord)
        sample_point_coord_names.append(coord.name())
        value = np.array(value, ndmin=1)
        coord_values.append(value)

    coord_point_lens = np.array([len(value) for value in coord_values])
    if not np.all(coord_point_lens == coord_point_lens[0]):
        msg = "All coordinates must have the same number of sample points."
        raise ValueError(msg)

    coord_values = np.array(coord_values)

    # Which dims are we sampling?
    sample_dims = set()
    for coord in sample_point_coords:
        for dim in cube.coord_dims(coord):
            sample_dims.add(dim)
    sample_dims = sorted(list(sample_dims))

    # Extract a sub cube that lives in just the sampling space.
    sample_space_slice = [0] * cube.ndim
    for sample_dim in sample_dims:
        sample_space_slice[sample_dim] = slice(None, None)
    sample_space_slice = tuple(sample_space_slice)
    sample_space_cube = cube[sample_space_slice]

    # Just the sampling coords.
    for coord in sample_space_cube.coords():
        if not coord.name() in sample_point_coord_names:
            sample_space_cube.remove_coord(coord)

    # Order the sample point coords according to the sample space cube coords.
    sample_space_coord_names = [
        coord.name() for coord in sample_space_cube.coords()
    ]
    new_order = [
        sample_space_coord_names.index(name)
        for name in sample_point_coord_names
    ]
    coord_values = np.array([coord_values[i] for i in new_order])
    sample_point_coord_names = [sample_point_coord_names[i] for i in new_order]

    sample_space_coords = (
        sample_space_cube.dim_coords + sample_space_cube.aux_coords
    )
    sample_space_coords_and_dims = [
        (coord, sample_space_cube.coord_dims(coord))
        for coord in sample_space_coords
    ]

    if cache is not None and cube in cache:
        kdtree = cache[cube]
    else:
        # Create a "sample space position" for each
        # `datum.sample_space_data_positions[coord_index][datum_index]`.
        sample_space_data_positions = np.empty(
            (len(sample_space_coords_and_dims), sample_space_cube.data.size),
            dtype=float,
        )
        for d, ndi in enumerate(np.ndindex(sample_space_cube.data.shape)):
            for c, (coord, coord_dims) in enumerate(
                sample_space_coords_and_dims
            ):
                # Index of this datum along this coordinate (could be nD).
                if coord_dims:
                    keys = tuple(ndi[ind] for ind in coord_dims)
                else:
                    keys = slice(None, None)
                # Position of this datum along this coordinate.
                sample_space_data_positions[c][d] = coord.points[keys]

        # Convert to cartesian coordinates. Flatten for kdtree compatibility.
        cartesian_space_data_coords = _cartesian_sample_points(
            sample_space_data_positions, sample_point_coord_names
        )

        # Create a kdtree for the nearest-distance lookup to these 3d points.
        kdtree = cKDTree(cartesian_space_data_coords)
        # This can find the nearest datum point to any given target point,
        # which is the goal of this function.

    # Update cache.
    if cache is not None:
        cache[cube] = kdtree

    # Convert the sample points to cartesian (3d) coords.
    # If there is no latlon within the coordinate there will be no change.
    # Otherwise, geographic latlon is replaced with cartesian xyz.
    cartesian_sample_points = _cartesian_sample_points(
        coord_values, sample_point_coord_names
    )

    # Use kdtree to get the nearest sourcepoint index for each target point.
    _, datum_index_lists = kdtree.query(cartesian_sample_points)

    # Convert flat indices back into multidimensional sample-space indices.
    sample_space_dimension_indices = np.unravel_index(
        datum_index_lists, sample_space_cube.data.shape
    )
    # Convert this from "pointwise list of index arrays for each dimension",
    # to "list of cube indices for each point".
    sample_space_ndis = np.array(sample_space_dimension_indices).transpose()

    # For the returned result, we must convert these indices into the source
    # (sample-space) cube, to equivalent indices into the target 'cube'.

    # Make a result array: (cube.ndim * <index>), per sample point.
    n_points = coord_values.shape[-1]
    main_cube_slices = np.empty((n_points, cube.ndim), dtype=object)
    # Initialise so all unused indices are ":".
    main_cube_slices[:] = slice(None)

    # Move result indices according to the source (sample) and target (cube)
    # dimension mappings.
    for sample_coord, sample_coord_dims in sample_space_coords_and_dims:
        # Find the coord in the main cube
        main_coord = cube.coord(sample_coord.name())
        main_coord_dims = cube.coord_dims(main_coord)
        # Fill nearest-point data indices for each coord dimension.
        for sample_i, main_i in zip(sample_coord_dims, main_coord_dims):
            main_cube_slices[:, main_i] = sample_space_ndis[:, sample_i]

    # Return as a list of **tuples** : required for correct indexing usage.
    result = [tuple(inds) for inds in main_cube_slices]
    return result


class UnstructuredNearestNeigbourRegridder:
    """
    Encapsulate the operation of :meth:`iris.analysis.trajectory.interpolate`
    with given source and target grids.

    This is the type used by the :class:`~iris.analysis.UnstructuredNearest`
    regridding scheme.

    """

    # TODO: cache the necessary bits of the operation so re-use can actually
    # be more efficient.
    def __init__(self, src_cube, target_grid_cube):
        """
        A nearest-neighbour regridder to perform regridding from the source
        grid to the target grid.

        This can then be applied to any source data with the same structure as
        the original 'src_cube'.

        Args:

        * src_cube:
            The :class:`~iris.cube.Cube` defining the source grid.
            The X and Y coordinates can have any shape, but must be mapped over
            the same cube dimensions.

        * target_grid_cube:
            A :class:`~iris.cube.Cube`, whose X and Y coordinates specify a
            desired target grid.
            The X and Y coordinates must be one-dimensional dimension
            coordinates, mapped to different dimensions.
            All other cube components are ignored.

        Returns:
            regridder : (object)

            A callable object with the interface:
                `result_cube = regridder(data)`

            where `data` is a cube with the same grid as the original
            `src_cube`, that is to be regridded to the `target_grid_cube`.

        .. Note::

            For latitude-longitude coordinates, the nearest-neighbour distances
            are computed on the sphere, otherwise flat Euclidean distances are
            used.

            The source and target X and Y coordinates must all have the same
            coordinate system, which may also be None.
            If any X and Y coordinates are latitudes or longitudes, they *all*
            must be.  Otherwise, the corresponding X and Y coordinates must
            have the same units in the source and grid cubes.

        """
        # Make a copy of the source cube, so we can convert coordinate units.
        src_cube = src_cube.copy()

        # Snapshot the target grid and check it is a "normal" grid.
        tgt_x_coord, tgt_y_coord = snapshot_grid(target_grid_cube)

        # Check that the source has unique X and Y coords over common dims.
        if not src_cube.coords(axis="x") or not src_cube.coords(axis="y"):
            msg = "Source cube must have X- and Y-axis coordinates."
            raise ValueError(msg)
        src_x_coord = src_cube.coord(axis="x")
        src_y_coord = src_cube.coord(axis="y")
        if src_cube.coord_dims(src_x_coord) != src_cube.coord_dims(
            src_y_coord
        ):
            msg = (
                "Source cube X and Y coordinates must have the same "
                "cube dimensions."
            )
            raise ValueError(msg)

        # Record *copies* of the original grid coords, in the desired
        # dimension order.
        # This lets us convert the actual ones in use to units of "degrees".
        self.src_grid_coords = [src_y_coord.copy(), src_x_coord.copy()]
        self.tgt_grid_coords = [tgt_y_coord.copy(), tgt_x_coord.copy()]

        # Check that all XY coords have suitable coordinate systems and units.
        coords_all = [src_x_coord, src_y_coord, tgt_x_coord, tgt_y_coord]
        cs = coords_all[0].coord_system
        if not all(coord.coord_system == cs for coord in coords_all):
            msg = (
                "Source and target cube X and Y coordinates must all have "
                "the same coordinate system."
            )
            raise ValueError(msg)

        # Check *all* X and Y coords are lats+lons, if any are.
        latlons = [
            "latitude" in coord.name() or "longitude" in coord.name()
            for coord in coords_all
        ]
        if any(latlons) and not all(latlons):
            msg = (
                "If any X and Y coordinates are latitudes/longitudes, "
                "then they all must be."
            )
            raise ValueError(msg)

        self.grid_is_latlon = any(latlons)
        if self.grid_is_latlon:
            # Convert all XY coordinates to units of "degrees".
            # N.B. already copied the target grid, so the result matches that.
            for coord in coords_all:
                try:
                    coord.convert_units("degrees")
                except ValueError:
                    msg = (
                        "Coordinate {!r} has units of {!r}, which does not "
                        'convert to "degrees".'
                    )
                    raise ValueError(
                        msg.format(coord.name(), str(coord.units))
                    )
        else:
            # Check that source and target have the same X and Y units.
            if (
                src_x_coord.units != tgt_x_coord.units
                or src_y_coord.units != tgt_y_coord.units
            ):
                msg = (
                    "Source and target cube X and Y coordinates must "
                    "have the same units."
                )
                raise ValueError(msg)

        # Record the resulting grid shape.
        self.tgt_grid_shape = tgt_y_coord.shape + tgt_x_coord.shape

        # Calculate sample points as 2d arrays, like broadcast (NY,1)*(1,NX).
        x_2d, y_2d = _meshgrid(tgt_x_coord.points, tgt_y_coord.points)
        # Cast as a "trajectory", to suit the method used.
        self.trajectory = (
            (tgt_x_coord.name(), x_2d.flatten()),
            (tgt_y_coord.name(), y_2d.flatten()),
        )

    def __call__(self, src_cube):
        # Check the source cube X and Y coords match the original.
        # Note: for now, this is sufficient to ensure a valid trajectory
        # interpolation, but if in future we save + re-use the cache context
        # for the 'interpolate' call, we may need more checks here.

        # Check the given cube against the original.
        x_cos = src_cube.coords(axis="x")
        y_cos = src_cube.coords(axis="y")
        if (
            not x_cos
            or not y_cos
            or y_cos != [self.src_grid_coords[0]]
            or x_cos != [self.src_grid_coords[1]]
        ):
            msg = (
                "The given cube is not defined on the same source "
                "grid as this regridder."
            )
            raise ValueError(msg)

        # Convert source XY coordinates to degrees if required.
        if self.grid_is_latlon:
            src_cube = src_cube.copy()
            src_cube.coord(axis="x").convert_units("degrees")
            src_cube.coord(axis="y").convert_units("degrees")

        # Get the basic interpolated results.
        result_trajectory_cube = interpolate(
            src_cube, self.trajectory, method="nearest"
        )

        # Reconstruct this as a cube "like" the source data.
        # TODO: handle all aux-coords, cell measures ??

        # The shape is that of the basic result, minus the trajectory (last)
        # dimension, plus the target grid dimensions.
        target_shape = result_trajectory_cube.shape[:-1] + self.tgt_grid_shape
        data_2d_x_and_y = result_trajectory_cube.data.reshape(target_shape)

        # Make a new result cube with the reshaped data.
        result_cube = iris.cube.Cube(data_2d_x_and_y)
        result_cube.metadata = src_cube.metadata

        # Copy all the coords from the trajectory result.
        i_trajectory_dim = result_trajectory_cube.ndim - 1
        for coord in result_trajectory_cube.dim_coords:
            dims = result_trajectory_cube.coord_dims(coord)
            if i_trajectory_dim not in dims:
                result_cube.add_dim_coord(coord.copy(), dims)
        for coord in result_trajectory_cube.aux_coords:
            dims = result_trajectory_cube.coord_dims(coord)
            if i_trajectory_dim not in dims:
                result_cube.add_aux_coord(coord.copy(), dims)

        # Add the X+Y grid coords from the grid cube, mapped to the new Y and X
        # dimensions, i.e. the last 2.
        for i_dim, coord in enumerate(self.tgt_grid_coords):
            result_cube.add_dim_coord(coord.copy(), i_dim + i_trajectory_dim)

        return result_cube



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for :class:`iris.analysis.AreaWeighted`."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock

from iris.analysis import AreaWeighted


class Test(tests.IrisTest):
    def check_call(self, mdtol=None):
        # Check that `iris.analysis.AreaWeighted` correctly calls an
        # `iris.analysis._area_weighted.AreaWeightedRegridder` object.
        if mdtol is None:
            area_weighted = AreaWeighted()
            mdtol = 1
        else:
            area_weighted = AreaWeighted(mdtol=mdtol)
        self.assertEqual(area_weighted.mdtol, mdtol)

        with mock.patch(
            "iris.analysis.AreaWeightedRegridder",
            return_value=mock.sentinel.regridder,
        ) as awr:
            regridder = area_weighted.regridder(
                mock.sentinel.src, mock.sentinel.target
            )

        awr.assert_called_once_with(
            mock.sentinel.src, mock.sentinel.target, mdtol=mdtol
        )
        self.assertIs(regridder, mock.sentinel.regridder)

    def test_default(self):
        self.check_call()

    def test_specified_mdtol(self):
        self.check_call(0.5)

    def test_invalid_high_mdtol(self):
        msg = "mdtol must be in range 0 - 1"
        with self.assertRaisesRegex(ValueError, msg):
            AreaWeighted(mdtol=1.2)

    def test_invalid_low_mdtol(self):
        msg = "mdtol must be in range 0 - 1"
        with self.assertRaisesRegex(ValueError, msg):
            AreaWeighted(mdtol=-0.2)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Test function :func:`iris.util.rolling_window`."""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

import numpy as np
import numpy.ma as ma

from iris.util import rolling_window


class Test_rolling_window(tests.IrisTest):
    def test_1d(self):
        # 1-d array input
        a = np.array([0, 1, 2, 3, 4], dtype=np.int32)
        expected_result = np.array(
            [[0, 1], [1, 2], [2, 3], [3, 4]], dtype=np.int32
        )
        result = rolling_window(a, window=2)
        self.assertArrayEqual(result, expected_result)

    def test_2d(self):
        # 2-d array input
        a = np.array([[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]], dtype=np.int32)
        expected_result = np.array(
            [
                [[0, 1, 2], [1, 2, 3], [2, 3, 4]],
                [[5, 6, 7], [6, 7, 8], [7, 8, 9]],
            ],
            dtype=np.int32,
        )
        result = rolling_window(a, window=3, axis=1)
        self.assertArrayEqual(result, expected_result)

    def test_1d_masked(self):
        # 1-d masked array input
        a = ma.array([0, 1, 2, 3, 4], mask=[0, 0, 1, 0, 0], dtype=np.int32)
        expected_result = ma.array(
            [[0, 1], [1, 2], [2, 3], [3, 4]],
            mask=[[0, 0], [0, 1], [1, 0], [0, 0]],
            dtype=np.int32,
        )
        result = rolling_window(a, window=2)
        self.assertMaskedArrayEqual(result, expected_result)

    def test_2d_masked(self):
        # 2-d masked array input
        a = ma.array(
            [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]],
            mask=[[0, 0, 1, 0, 0], [1, 0, 1, 0, 0]],
            dtype=np.int32,
        )
        expected_result = ma.array(
            [
                [[0, 1, 2], [1, 2, 3], [2, 3, 4]],
                [[5, 6, 7], [6, 7, 8], [7, 8, 9]],
            ],
            mask=[
                [[0, 0, 1], [0, 1, 0], [1, 0, 0]],
                [[1, 0, 1], [0, 1, 0], [1, 0, 0]],
            ],
            dtype=np.int32,
        )
        result = rolling_window(a, window=3, axis=1)
        self.assertMaskedArrayEqual(result, expected_result)

    def test_degenerate_mask(self):
        a = ma.array([[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]], dtype=np.int32)
        expected_result = ma.array(
            [
                [[0, 1, 2], [1, 2, 3], [2, 3, 4]],
                [[5, 6, 7], [6, 7, 8], [7, 8, 9]],
            ],
            mask=[
                [[0, 0, 0], [0, 0, 0], [0, 0, 0]],
                [[0, 0, 0], [0, 0, 0], [0, 0, 0]],
            ],
            dtype=np.int32,
        )
        result = rolling_window(a, window=3, axis=1)
        self.assertMaskedArrayEqual(result, expected_result)

    def test_step(self):
        # step should control how far apart consecutive windows are
        a = np.array([[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]], dtype=np.int32)
        expected_result = np.array(
            [[[0, 1, 2], [2, 3, 4]], [[5, 6, 7], [7, 8, 9]]], dtype=np.int32
        )
        result = rolling_window(a, window=3, step=2, axis=1)
        self.assertArrayEqual(result, expected_result)

    def test_window_too_short(self):
        # raise an error if the window length is less than 1
        a = np.empty([5])
        with self.assertRaises(ValueError):
            rolling_window(a, window=0)

    def test_window_too_long(self):
        # raise an error if the window length is longer than the
        # corresponding array dimension
        a = np.empty([7, 5])
        with self.assertRaises(ValueError):
            rolling_window(a, window=6, axis=1)

    def test_invalid_step(self):
        # raise an error if the step between windows is less than 1
        a = np.empty([5])
        with self.assertRaises(ValueError):
            rolling_window(a, step=0)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

# import iris tests first so that some things can be initialised before importing anything else
import iris.tests as tests

import iris
import iris.fileformats.pp
import iris.io
import iris.util
import iris.tests.stock


class TestPPStash(tests.IrisTest):
    @tests.skip_data
    def test_cube_attributes(self):
        cube = tests.stock.simple_pp()
        self.assertEqual("m01s16i203", cube.attributes["STASH"])
        self.assertNotEqual("m01s16i999", cube.attributes["STASH"])
        self.assertEqual(cube.attributes["STASH"], "m01s16i203")
        self.assertNotEqual(cube.attributes["STASH"], "m01s16i999")

    @tests.skip_data
    def test_ppfield(self):
        data_path = tests.get_data_path(("PP", "simple_pp", "global.pp"))
        pps = iris.fileformats.pp.load(data_path)
        for pp in pps:
            self.assertEqual("m01s16i203", pp.stash)
            self.assertNotEqual("m01s16i999", pp.stash)
            self.assertEqual(pp.stash, "m01s16i203")
            self.assertNotEqual(pp.stash, "m01s16i999")

    def test_stash_against_stash(self):
        self.assertEqual(
            iris.fileformats.pp.STASH(1, 2, 3),
            iris.fileformats.pp.STASH(1, 2, 3),
        )
        self.assertNotEqual(
            iris.fileformats.pp.STASH(1, 2, 3),
            iris.fileformats.pp.STASH(2, 3, 4),
        )

    def test_stash_against_str(self):
        self.assertEqual(iris.fileformats.pp.STASH(1, 2, 3), "m01s02i003")
        self.assertEqual("m01s02i003", iris.fileformats.pp.STASH(1, 2, 3))
        self.assertNotEqual(iris.fileformats.pp.STASH(1, 2, 3), "m02s03i004")
        self.assertNotEqual("m02s03i004", iris.fileformats.pp.STASH(1, 2, 3))

    def test_irregular_stash_str(self):
        self.assertEqual(
            iris.fileformats.pp.STASH(1, 2, 3), "m01s02i0000000003"
        )
        self.assertEqual(iris.fileformats.pp.STASH(1, 2, 3), "m01s02i3")
        self.assertEqual(iris.fileformats.pp.STASH(1, 2, 3), "m01s2i3")
        self.assertEqual(iris.fileformats.pp.STASH(1, 2, 3), "m1s2i3")

        self.assertEqual(
            "m01s02i0000000003", iris.fileformats.pp.STASH(1, 2, 3)
        )
        self.assertEqual("m01s02i3", iris.fileformats.pp.STASH(1, 2, 3))
        self.assertEqual("m01s2i3", iris.fileformats.pp.STASH(1, 2, 3))
        self.assertEqual("m1s2i3", iris.fileformats.pp.STASH(1, 2, 3))

        self.assertNotEqual(
            iris.fileformats.pp.STASH(2, 3, 4), "m01s02i0000000003"
        )
        self.assertNotEqual(iris.fileformats.pp.STASH(2, 3, 4), "m01s02i3")
        self.assertNotEqual(iris.fileformats.pp.STASH(2, 3, 4), "m01s2i3")
        self.assertNotEqual(iris.fileformats.pp.STASH(2, 3, 4), "m1s2i3")

        self.assertNotEqual(
            "m01s02i0000000003", iris.fileformats.pp.STASH(2, 3, 4)
        )
        self.assertNotEqual("m01s02i3", iris.fileformats.pp.STASH(2, 3, 4))
        self.assertNotEqual("m01s2i3", iris.fileformats.pp.STASH(2, 3, 4))
        self.assertNotEqual("m1s2i3", iris.fileformats.pp.STASH(2, 3, 4))

        self.assertEqual(
            iris.fileformats.pp.STASH.from_msi("M01s02i003"), "m01s02i003"
        )
        self.assertEqual(
            "m01s02i003", iris.fileformats.pp.STASH.from_msi("M01s02i003")
        )

    def test_illegal_stash_str_range(self):

        self.assertEqual(iris.fileformats.pp.STASH(0, 2, 3), "m??s02i003")
        self.assertNotEqual(iris.fileformats.pp.STASH(0, 2, 3), "m01s02i003")

        self.assertEqual("m??s02i003", iris.fileformats.pp.STASH(0, 2, 3))
        self.assertNotEqual("m01s02i003", iris.fileformats.pp.STASH(0, 2, 3))

        self.assertEqual(iris.fileformats.pp.STASH(0, 2, 3), "m??s02i003")
        self.assertEqual(iris.fileformats.pp.STASH(0, 2, 3), "m00s02i003")
        self.assertEqual("m??s02i003", iris.fileformats.pp.STASH(0, 2, 3))
        self.assertEqual("m00s02i003", iris.fileformats.pp.STASH(0, 2, 3))

        self.assertEqual(iris.fileformats.pp.STASH(100, 2, 3), "m??s02i003")
        self.assertEqual(iris.fileformats.pp.STASH(100, 2, 3), "m100s02i003")
        self.assertEqual("m??s02i003", iris.fileformats.pp.STASH(100, 2, 3))
        self.assertEqual("m100s02i003", iris.fileformats.pp.STASH(100, 2, 3))

    def test_illegal_stash_stash_range(self):
        self.assertEqual(
            iris.fileformats.pp.STASH(0, 2, 3),
            iris.fileformats.pp.STASH(0, 2, 3),
        )
        self.assertEqual(
            iris.fileformats.pp.STASH(100, 2, 3),
            iris.fileformats.pp.STASH(100, 2, 3),
        )
        self.assertEqual(
            iris.fileformats.pp.STASH(100, 2, 3),
            iris.fileformats.pp.STASH(999, 2, 3),
        )

    def test_illegal_stash_format(self):
        test_values = (
            ("abc", (1, 2, 3)),
            ("mlotstmin", (1, 2, 3)),
            ("m01s02003", (1, 2, 3)),
        )

        for (test_value, reference) in test_values:
            msg = "Expected STASH code .* {!r}".format(test_value)
            with self.assertRaisesRegex(ValueError, msg):
                test_value == iris.fileformats.pp.STASH(*reference)
            with self.assertRaisesRegex(ValueError, msg):
                iris.fileformats.pp.STASH(*reference) == test_value

    def test_illegal_stash_type(self):
        test_values = (
            (102003, "m01s02i003"),
            (["m01s02i003"], "m01s02i003"),
        )

        for (test_value, reference) in test_values:
            msg = "Expected STASH code .* {!r}".format(test_value)
            with self.assertRaisesRegex(TypeError, msg):
                iris.fileformats.pp.STASH.from_msi(test_value) == reference
            with self.assertRaisesRegex(TypeError, msg):
                reference == iris.fileformats.pp.STASH.from_msi(test_value)

    def test_stash_lbuser(self):
        stash = iris.fileformats.pp.STASH(2, 32, 456)
        self.assertEqual(stash.lbuser6(), 2)
        self.assertEqual(stash.lbuser3(), 32456)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for :class:`iris.fileformats.rules.Loader`."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock

from iris.fileformats.rules import Loader


class Test___init__(tests.IrisTest):
    def test_normal(self):
        with mock.patch("warnings.warn") as warn:
            loader = Loader(
                mock.sentinel.GEN_FUNC,
                mock.sentinel.GEN_FUNC_KWARGS,
                mock.sentinel.CONVERTER,
            )
        self.assertEqual(warn.call_count, 0)
        self.assertIs(loader.field_generator, mock.sentinel.GEN_FUNC)
        self.assertIs(
            loader.field_generator_kwargs, mock.sentinel.GEN_FUNC_KWARGS
        )
        self.assertIs(loader.converter, mock.sentinel.CONVERTER)

    def test_normal_with_explicit_none(self):
        with mock.patch("warnings.warn") as warn:
            loader = Loader(
                mock.sentinel.GEN_FUNC,
                mock.sentinel.GEN_FUNC_KWARGS,
                mock.sentinel.CONVERTER,
            )
        self.assertEqual(warn.call_count, 0)
        self.assertIs(loader.field_generator, mock.sentinel.GEN_FUNC)
        self.assertIs(
            loader.field_generator_kwargs, mock.sentinel.GEN_FUNC_KWARGS
        )
        self.assertIs(loader.converter, mock.sentinel.CONVERTER)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Test function :func:`iris.util._is_circular`."""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

import numpy as np

from iris.util import _is_circular


class Test(tests.IrisTest):
    def test_simple(self):
        data = np.arange(12) * 30
        self.assertTrue(_is_circular(data, 360))

    def test_negative_diff(self):
        data = (np.arange(96) * -3.749998) + 3.56249908e02
        self.assertTrue(_is_circular(data, 360))


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Provides UK Met Office Fields File (FF) format specific capabilities.

"""

import os
import warnings

import numpy as np

from iris.exceptions import NotYetImplementedError
from iris.fileformats._ff_cross_references import STASH_TRANS
from . import pp


IMDI = -32768

FF_HEADER_DEPTH = 256  # In words (64-bit).
DEFAULT_FF_WORD_DEPTH = 8  # In bytes.

# UM marker to signify empty lookup table entry.
_FF_LOOKUP_TABLE_TERMINATE = -99

# UM FieldsFile fixed length header names and positions.
UM_FIXED_LENGTH_HEADER = [
    ("data_set_format_version", (1,)),
    ("sub_model", (2,)),
    ("vert_coord_type", (3,)),
    ("horiz_grid_type", (4,)),
    ("dataset_type", (5,)),
    ("run_identifier", (6,)),
    ("experiment_number", (7,)),
    ("calendar", (8,)),
    ("grid_staggering", (9,)),
    ("time_type", (10,)),
    ("projection_number", (11,)),
    ("model_version", (12,)),
    ("obs_file_type", (14,)),
    ("last_fieldop_type", (15,)),
    ("first_validity_time", (21, 22, 23, 24, 25, 26, 27)),
    ("last_validity_time", (28, 29, 30, 31, 32, 33, 34)),
    ("misc_validity_time", (35, 36, 37, 38, 39, 40, 41)),
    ("integer_constants", (100, 101)),
    ("real_constants", (105, 106)),
    ("level_dependent_constants", (110, 111, 112)),
    ("row_dependent_constants", (115, 116, 117)),
    ("column_dependent_constants", (120, 121, 122)),
    ("fields_of_constants", (125, 126, 127)),
    ("extra_constants", (130, 131)),
    ("temp_historyfile", (135, 136)),
    ("compressed_field_index1", (140, 141)),
    ("compressed_field_index2", (142, 143)),
    ("compressed_field_index3", (144, 145)),
    ("lookup_table", (150, 151, 152)),
    ("total_prognostic_fields", (153,)),
    ("data", (160, 161, 162)),
]

# Offset value to convert from UM_FIXED_LENGTH_HEADER positions to
# FF_HEADER offsets.
UM_TO_FF_HEADER_OFFSET = 1
# Offset the UM_FIXED_LENGTH_HEADER positions to FF_HEADER offsets.
FF_HEADER = [
    (name, tuple(position - UM_TO_FF_HEADER_OFFSET for position in positions))
    for name, positions in UM_FIXED_LENGTH_HEADER
]

# UM FieldsFile fixed length header pointer names.
_FF_HEADER_POINTERS = [
    "integer_constants",
    "real_constants",
    "level_dependent_constants",
    "row_dependent_constants",
    "column_dependent_constants",
    "fields_of_constants",
    "extra_constants",
    "temp_historyfile",
    "compressed_field_index1",
    "compressed_field_index2",
    "compressed_field_index3",
    "lookup_table",
    "data",
]

_LBUSER_DTYPE_LOOKUP = {
    1: ">f{word_depth}",
    2: ">i{word_depth}",
    3: ">i{word_depth}",
    "default": ">f{word_depth}",
}

#: Codes used in STASH_GRID which indicate the x coordinate is on the
#: edge of the cell.
X_COORD_U_GRID = (11, 18, 27)

#: Codes used in STASH_GRID which indicate the y coordinate is on the
#: edge of the cell.
Y_COORD_V_GRID = (11, 19, 28)

#: Grid codes found in the STASH master which are currently known to be
#: handled correctly. A warning is issued if a grid is found which is not
#: handled.
HANDLED_GRIDS = (1, 2, 3, 4, 5, 21, 26, 29) + X_COORD_U_GRID + Y_COORD_V_GRID

# REAL constants header names as described by UM documentation paper F3.
# NB. These are zero-based indices as opposed to the one-based indices
# used in F3.
REAL_EW_SPACING = 0
REAL_NS_SPACING = 1
REAL_FIRST_LAT = 2
REAL_FIRST_LON = 3
REAL_POLE_LAT = 4
REAL_POLE_LON = 5


class Grid:
    """
    An abstract class representing the default/file-level grid
    definition for a FieldsFile.

    """

    def __init__(
        self,
        column_dependent_constants,
        row_dependent_constants,
        real_constants,
        horiz_grid_type,
    ):
        """
        Create a Grid from the relevant sections of the FFHeader.

        Args:

        * column_dependent_constants (numpy.ndarray):
            The `column_dependent_constants` from a FFHeader.

        * row_dependent_constants (numpy.ndarray):
            The `row_dependent_constants` from a FFHeader.

        * real_constants (numpy.ndarray):
            The `real_constants` from a FFHeader.

        * horiz_grid_type (integer):
            `horiz_grid_type` from a FFHeader.

        """
        self.column_dependent_constants = column_dependent_constants
        self.row_dependent_constants = row_dependent_constants
        self.ew_spacing = real_constants[REAL_EW_SPACING]
        self.ns_spacing = real_constants[REAL_NS_SPACING]
        self.first_lat = real_constants[REAL_FIRST_LAT]
        self.first_lon = real_constants[REAL_FIRST_LON]
        self.pole_lat = real_constants[REAL_POLE_LAT]
        self.pole_lon = real_constants[REAL_POLE_LON]
        self.horiz_grid_type = horiz_grid_type

    def _x_vectors(self, subgrid):
        # Abstract method to return the X vector for the given sub-grid.
        raise NotImplementedError()

    def _y_vectors(self, subgrid):
        # Abstract method to return the X vector for the given sub-grid.
        raise NotImplementedError()

    def regular_x(self, subgrid):
        # Abstract method to return BZX, BDX for the given sub-grid.
        raise NotImplementedError()

    def regular_y(self, subgrid):
        # Abstract method to return BZY, BDY for the given sub-grid.
        raise NotImplementedError()

    def vectors(self, subgrid):
        """
        Return the X and Y coordinate vectors for the given sub-grid of
        this grid.

        Args:

        * subgrid (integer):
            A "grid type code" as described in UM documentation paper C4.

        Returns:
            A 2-tuple of X-vector, Y-vector.

        """
        x_p, x_u = self._x_vectors()
        y_p, y_v = self._y_vectors()
        x = x_p
        y = y_p
        if subgrid in X_COORD_U_GRID:
            x = x_u
        if subgrid in Y_COORD_V_GRID:
            y = y_v
        return x, y


class ArakawaC(Grid):
    """
    An abstract class representing an Arakawa C-grid.

    """

    def _x_vectors(self):
        x_p, x_u = None, None
        if self.column_dependent_constants is not None:
            x_p = self.column_dependent_constants[:, 0]
            if self.column_dependent_constants.shape[1] == 2:
                # Wrap around for global field
                if self.horiz_grid_type == 0:
                    x_u = self.column_dependent_constants[:-1, 1]
                else:
                    x_u = self.column_dependent_constants[:, 1]
        return x_p, x_u

    def regular_x(self, subgrid):
        """
        Return the "zeroth" value and step for the X coordinate on the
        given sub-grid of this grid.

        Args:

        * subgrid (integer):
            A "grid type code" as described in UM documentation paper C4.

        Returns:
            A 2-tuple of BZX, BDX.

        """
        bdx = self.ew_spacing
        bzx = self.first_lon - bdx
        if subgrid in X_COORD_U_GRID:
            bzx += 0.5 * bdx
        return bzx, bdx

    def regular_y(self, subgrid):
        """
        Return the "zeroth" value and step for the Y coordinate on the
        given sub-grid of this grid.

        Args:

        * subgrid (integer):
            A "grid type code" as described in UM documentation paper C4.

        Returns:
            A 2-tuple of BZY, BDY.

        """
        bdy = self.ns_spacing
        bzy = self.first_lat - bdy
        if subgrid in Y_COORD_V_GRID:
            bzy += self._v_offset * bdy
        return bzy, bdy


class NewDynamics(ArakawaC):
    """
    An Arakawa C-grid as used by UM New Dynamics.

    The theta and u points are at the poles.

    """

    _v_offset = 0.5

    def _y_vectors(self):
        y_p, y_v = None, None
        if self.row_dependent_constants is not None:
            y_p = self.row_dependent_constants[:, 0]
            if self.row_dependent_constants.shape[1] == 2:
                y_v = self.row_dependent_constants[:-1, 1]
        return y_p, y_v


class ENDGame(ArakawaC):
    """
    An Arakawa C-grid as used by UM ENDGame.

    The v points are at the poles.

    """

    _v_offset = -0.5

    def _y_vectors(self):
        y_p, y_v = None, None
        if self.row_dependent_constants is not None:
            y_p = self.row_dependent_constants[:-1, 0]
            if self.row_dependent_constants.shape[1] == 2:
                y_v = self.row_dependent_constants[:, 1]
        return y_p, y_v


class FFHeader:
    """
    A class to represent the FIXED_LENGTH_HEADER section of a FieldsFile.

    """

    GRID_STAGGERING_CLASS = {3: NewDynamics, 6: ENDGame}

    def __init__(self, filename, word_depth=DEFAULT_FF_WORD_DEPTH):
        """
        Create a FieldsFile header instance by reading the
        FIXED_LENGTH_HEADER section of the FieldsFile, making the names
        defined in FF_HEADER available as attributes of a FFHeader instance.

        Args:

        * filename (string):
            Specify the name of the FieldsFile.

        Returns:
            FFHeader object.

        """

        #: File name of the FieldsFile.
        self.ff_filename = filename
        self._word_depth = word_depth

        # Read the FF header data
        with open(filename, "rb") as ff_file:
            # typically 64-bit words (aka. int64 or ">i8")
            header_data = _parse_binary_stream(
                ff_file,
                dtype=">i{0}".format(word_depth),
                count=FF_HEADER_DEPTH,
            )
            header_data = tuple(header_data)
            # Create FF instance attributes
            for name, offsets in FF_HEADER:
                if len(offsets) == 1:
                    value = header_data[offsets[0]]
                else:
                    value = header_data[offsets[0] : offsets[-1] + 1]
                setattr(self, name, value)

            # Turn the pointer values into real arrays.
            for elem in _FF_HEADER_POINTERS:
                if elem not in ["data", "lookup_table"]:
                    if self._attribute_is_pointer_and_needs_addressing(elem):
                        addr = getattr(self, elem)
                        ff_file.seek((addr[0] - 1) * word_depth, os.SEEK_SET)
                        if len(addr) == 2:
                            if elem == "integer_constants":
                                res = _parse_binary_stream(
                                    ff_file,
                                    dtype=">i{0}".format(word_depth),
                                    count=addr[1],
                                )
                            else:
                                res = _parse_binary_stream(
                                    ff_file,
                                    dtype=">f{0}".format(word_depth),
                                    count=addr[1],
                                )
                        elif len(addr) == 3:
                            res = _parse_binary_stream(
                                ff_file,
                                dtype=">f{0}".format(word_depth),
                                count=addr[1] * addr[2],
                            )
                            res = res.reshape((addr[1], addr[2]), order="F")
                        else:
                            raise ValueError(
                                "ff header element {} is not"
                                "handled correctly".format(elem)
                            )
                    else:
                        res = None
                    setattr(self, elem, res)

    def __str__(self):
        attributes = []
        for name, _ in FF_HEADER:
            attributes.append("    {}: {}".format(name, getattr(self, name)))
        return "FF Header:\n" + "\n".join(attributes)

    def __repr__(self):
        return "{}({!r})".format(type(self).__name__, self.ff_filename)

    def _attribute_is_pointer_and_needs_addressing(self, name):
        if name in _FF_HEADER_POINTERS:
            attr = getattr(self, name)

            # Check that we haven't already addressed this pointer,
            # that the pointer is actually referenceable (i.e. >0)
            # and that the attribute is not marked as missing.
            is_referenceable = (
                isinstance(attr, tuple) and attr[0] > 0 and attr[0] != IMDI
            )
        else:
            msg = "{!r} object does not have pointer attribute {!r}"
            raise AttributeError(msg.format(self.__class__.__name__, name))
        return is_referenceable

    def shape(self, name):
        """
        Return the dimension shape of the FieldsFile FIXED_LENGTH_HEADER
        pointer attribute.

        Args:

        * name (string):
            Specify the name of the FIXED_LENGTH_HEADER attribute.

        Returns:
            Dimension tuple.

        """

        if name in _FF_HEADER_POINTERS:
            value = getattr(self, name)[1:]
        else:
            msg = "{!r} object does not have pointer address {!r}"
            raise AttributeError(msg.format(self.__class_.__name__, name))
        return value

    def grid(self):
        """Return the Grid definition for the FieldsFile."""
        grid_class = self.GRID_STAGGERING_CLASS.get(self.grid_staggering)
        if grid_class is None:
            grid_class = NewDynamics
            warnings.warn(
                "Staggered grid type: {} not currently interpreted, assuming "
                "standard C-grid".format(self.grid_staggering)
            )
        grid = grid_class(
            self.column_dependent_constants,
            self.row_dependent_constants,
            self.real_constants,
            self.horiz_grid_type,
        )
        return grid


class FF2PP:
    """
    A class to extract the individual PPFields from within a FieldsFile.

    """

    def __init__(
        self, filename, read_data=False, word_depth=DEFAULT_FF_WORD_DEPTH
    ):
        """
        Create a FieldsFile to Post Process instance that returns a generator
        of PPFields contained within the FieldsFile.

        Args:

        * filename (string):
            Specify the name of the FieldsFile.

        Kwargs:

        * read_data (boolean):
            Specify whether to read the associated PPField data within
            the FieldsFile.  Default value is False.

        Returns:
            PPField generator.

        For example::

            >>> for field in ff.FF2PP(filename):
            ...     print(field)

        """

        self._ff_header = FFHeader(filename, word_depth=word_depth)
        self._word_depth = word_depth
        self._filename = filename
        self._read_data = read_data

    def _payload(self, field):
        """Calculate the payload data depth (in bytes) and type."""
        lbpack_n1 = field.raw_lbpack % 10
        if lbpack_n1 == 0:
            word_depth = self._word_depth
            # Data payload is not packed.
            data_words = field.lblrec - field.lbext
            # Determine PP field 64-bit payload datatype.
            lookup = _LBUSER_DTYPE_LOOKUP
            dtype_template = lookup.get(field.lbuser[0], lookup["default"])
            dtype_name = dtype_template.format(word_depth=self._word_depth)
            data_type = np.dtype(dtype_name)
        else:
            word_depth = pp.PP_WORD_DEPTH
            # Data payload is packed.
            if lbpack_n1 == 1:
                # Data packed using WGDOS archive method.
                data_words = field.lbnrec * 2
            elif lbpack_n1 == 2:
                # Data packed using CRAY 32-bit method.
                data_words = field.lblrec - field.lbext
            else:
                msg = "PP fields with LBPACK of {} are not supported."
                raise NotYetImplementedError(msg.format(field.raw_lbpack))

            # Determine PP field payload datatype.
            lookup = pp.LBUSER_DTYPE_LOOKUP
            data_type = lookup.get(field.lbuser[0], lookup["default"])

        if field.boundary_packing is not None:
            if lbpack_n1 not in (0, 2):
                # Can't handle the usual packing methods with LBC data.
                raise ValueError(
                    "LBC data has LBPACK = {:d}, but packed LBC data is not "
                    "supported.".format(field.raw_lbpack)
                )
            # Adjust to packed data size, for LBC data.
            # NOTE: logic duplicates that in pp._data_bytes_to_shaped_array.
            pack_dims = field.boundary_packing
            boundary_height = pack_dims.y_halo + pack_dims.rim_width
            boundary_width = pack_dims.x_halo + pack_dims.rim_width
            y_height, x_width = field.lbrow, field.lbnpt
            mid_height = y_height - 2 * boundary_height
            data_words = (
                boundary_height * x_width * 2 + boundary_width * mid_height * 2
            )

        data_depth = data_words * word_depth
        return data_depth, data_type

    def _det_border(self, field_dim, halo_dim):
        # Update field coordinates for a variable resolution LBC file where
        # the resolution of the very edge (within the rim width) is assumed to
        # be same as the halo.
        def range_order(range1, range2, resolution):
            # Handles whether increasing/decreasing ranges.
            if np.sign(resolution) > 0:
                lower = range1
                upper = range2
            else:
                upper = range1
                lower = range2
            return lower, upper

        # Ensure that the resolution is the same on both edges.
        res_low = field_dim[1] - field_dim[0]
        res_high = field_dim[-1] - field_dim[-2]
        if not np.allclose(res_low, res_high):
            msg = (
                "The x or y coordinates of your boundary condition field "
                "may be incorrect, not having taken into account the "
                "boundary size."
            )
            warnings.warn(msg)
        else:
            range2 = field_dim[0] - res_low
            range1 = field_dim[0] - halo_dim * res_low
            lower, upper = range_order(range1, range2, res_low)
            extra_before = np.linspace(lower, upper, halo_dim)

            range1 = field_dim[-1] + res_high
            range2 = field_dim[-1] + halo_dim * res_high
            lower, upper = range_order(range1, range2, res_high)
            extra_after = np.linspace(lower, upper, halo_dim)

            field_dim = np.concatenate([extra_before, field_dim, extra_after])
        return field_dim

    def _adjust_field_for_lbc(self, field):
        """
        Make an LBC field look like a 'normal' field for rules processing.

        """
        # Set LBTIM to indicate the specific time encoding for LBCs,
        # i.e. t1=forecast, t2=reference
        lbtim_default = 11
        if field.lbtim not in (0, lbtim_default):
            raise ValueError(
                "LBC field has LBTIM of {:d}, expected only "
                "0 or {:d}.".format(field.lbtim, lbtim_default)
            )
        field.lbtim = lbtim_default

        # Set LBVC to indicate the specific height encoding for LBCs,
        # i.e. hybrid height layers.
        lbvc_default = 65
        if field.lbvc not in (0, lbvc_default):
            raise ValueError(
                "LBC field has LBVC of {:d}, expected only "
                "0 or {:d}.".format(field.lbvc, lbvc_default)
            )
        field.lbvc = lbvc_default
        # Specifying a vertical encoding scheme means a usable vertical
        # coordinate can be produced, because we also record the level
        # number in each result field:  Thus they are located, and can
        # be stacked, in a vertical dimension.
        # See per-layer loop (below).

        # Calculate field packing details.
        name_mapping = dict(
            rim_width=slice(4, 6), y_halo=slice(2, 4), x_halo=slice(0, 2)
        )
        boundary_packing = pp.SplittableInt(field.lbuser[2], name_mapping)
        # Store packing on the field, which affects how it gets the data.
        field.boundary_packing = boundary_packing
        # Fix the lbrow and lbnpt to be the actual size of the data
        # array, since the field is no longer a "boundary" fields file
        # field.
        # Note: The documentation states that lbrow (y) doesn't
        # contain the halo rows, but no such comment exists at UM v8.5
        # for lbnpt (x). Experimentation has shown that lbnpt also
        # excludes the halo size.
        field.lbrow += 2 * boundary_packing.y_halo
        field.lbnpt += 2 * boundary_packing.x_halo

        # Update the x and y coordinates for this field. Note: it may
        # be that this needs to update x and y also, but that is yet
        # to be confirmed.
        if field.bdx in (0, field.bmdi) or field.bdy in (0, field.bmdi):
            field.x = self._det_border(field.x, boundary_packing.x_halo)
            field.y = self._det_border(field.y, boundary_packing.y_halo)
        else:
            if field.bdy < 0:
                warnings.warn(
                    "The LBC has a bdy less than 0. No "
                    "case has previously been seen of "
                    "this, and the decompression may be "
                    "erroneous."
                )
            field.bzx -= field.bdx * boundary_packing.x_halo
            field.bzy -= field.bdy * boundary_packing.y_halo

    def _fields_over_all_levels(self, field):
        """
        Replicate the field over all model levels, setting LBLEV for each.

        This is appropriate for LBC data.
        Yields an iterator producing a sequence of distinct field objects.

        """
        n_all_levels = self._ff_header.level_dependent_constants.shape[0]
        levels_count = field.lbhem - 100
        if levels_count < 1:
            raise ValueError(
                "LBC field has LBHEM of {:d}, but this should be (100 "
                "+ levels-per-field-type), hence >= 101.".format(field.lbhem)
            )
        if levels_count > n_all_levels:
            raise ValueError(
                "LBC field has LBHEM of (100 + levels-per-field-type) "
                "= {:d}, but this is more than the total number of levels "
                "in the file = {:d}).".format(field.lbhem, n_all_levels)
            )

        for i_model_level in range(levels_count):
            # Make subsequent fields alike, but distinct.
            if i_model_level > 0:
                field = field.copy()
            # Provide the correct "model level" value.
            field.lblev = i_model_level
            # TODO: as LBC lookup headers cover multiple layers, they
            # contain no per-layer height values, which are all 0.
            # So the field's "height" coordinate will be useless.
            # It should be possible to fix this here, by calculating
            # per-layer Z and C values from the file header, and
            # setting blev/brlev/brsvd1 and bhlev/bhrlev/brsvd2 here,
            # but we don't yet do this.

            yield field

    def _extract_field(self):
        # FF table pointer initialisation based on FF LOOKUP table
        # configuration.

        lookup_table = self._ff_header.lookup_table
        table_index, table_entry_depth, table_count = lookup_table
        table_offset = (table_index - 1) * self._word_depth  # in bytes
        table_entry_depth = table_entry_depth * self._word_depth  # in bytes
        # Open the FF for processing.
        with open(self._ff_header.ff_filename, "rb") as ff_file:
            ff_file_seek = ff_file.seek

            is_boundary_packed = self._ff_header.dataset_type == 5

            grid = self._ff_header.grid()

            # Process each FF LOOKUP table entry.
            while table_count:
                table_count -= 1
                # Move file pointer to the start of the current FF LOOKUP
                # table entry.
                ff_file_seek(table_offset, os.SEEK_SET)

                # Read the current PP header entry from the FF LOOKUP table.
                header_longs = _parse_binary_stream(
                    ff_file,
                    dtype=">i{0}".format(self._word_depth),
                    count=pp.NUM_LONG_HEADERS,
                )
                # Check whether the current FF LOOKUP table entry is valid.
                if header_longs[0] == _FF_LOOKUP_TABLE_TERMINATE:
                    # There are no more FF LOOKUP table entries to read.
                    break
                header_floats = _parse_binary_stream(
                    ff_file,
                    dtype=">f{0}".format(self._word_depth),
                    count=pp.NUM_FLOAT_HEADERS,
                )
                header = tuple(header_longs) + tuple(header_floats)

                # Calculate next FF LOOKUP table entry.
                table_offset += table_entry_depth

                # Construct a PPField object and populate using the header_data
                # read from the current FF LOOKUP table.
                # (The PPField sub-class will depend on the header release
                # number.)
                # (Some Fields File fields are UM specific scratch spaces, with
                # no header release number, these will throw an exception from
                # the PP module and are skipped to enable reading of the file.
                try:
                    field = pp.make_pp_field(header)

                    # Fast stash look-up.
                    stash_s = field.lbuser[3] // 1000
                    stash_i = field.lbuser[3] % 1000
                    stash = "m{:02}s{:02}i{:03}".format(
                        field.lbuser[6], stash_s, stash_i
                    )
                    stash_entry = STASH_TRANS.get(stash, None)
                    if stash_entry is None:
                        subgrid = None
                    else:
                        subgrid = stash_entry.grid_code
                        if subgrid not in HANDLED_GRIDS:
                            warnings.warn(
                                "The stash code {} is on a grid {} "
                                "which has not been explicitly "
                                "handled by the fieldsfile loader."
                                " Assuming the data is on a P grid"
                                ".".format(stash, subgrid)
                            )

                    field.x, field.y = grid.vectors(subgrid)

                    # Use the per-file grid if no per-field metadata is
                    # available.
                    no_x = field.bzx in (0, field.bmdi) and field.x is None
                    no_y = field.bzy in (0, field.bmdi) and field.y is None
                    if no_x and no_y:
                        if subgrid is None:
                            msg = (
                                "The STASH code {0} was not found in the "
                                "STASH to grid type mapping. Picking the P "
                                "position as the cell type".format(stash)
                            )
                            warnings.warn(msg)
                        field.bzx, field.bdx = grid.regular_x(subgrid)
                        field.bzy, field.bdy = grid.regular_y(subgrid)
                        field.bplat = grid.pole_lat
                        field.bplon = grid.pole_lon
                    elif no_x or no_y:
                        warnings.warn(
                            "Partially missing X or Y coordinate values."
                        )

                    # Check for LBC fields.
                    is_boundary_packed = self._ff_header.dataset_type == 5
                    if is_boundary_packed:
                        # Apply adjustments specific to LBC data.
                        self._adjust_field_for_lbc(field)

                    # Calculate start address of the associated PP header data.
                    data_offset = field.lbegin * self._word_depth
                    # Determine PP field payload depth and type.
                    data_depth, data_type = self._payload(field)

                    # Produce (yield) output fields.
                    if is_boundary_packed:
                        fields = self._fields_over_all_levels(field)
                    else:
                        fields = [field]
                    for result_field in fields:
                        # Add a field data element.
                        if self._read_data:
                            # Read the actual bytes. This can then be converted
                            # to a numpy array at a higher level.
                            ff_file_seek(data_offset, os.SEEK_SET)
                            result_field.data = pp.LoadedArrayBytes(
                                ff_file.read(data_depth), data_type
                            )
                        else:
                            # Provide enough context to read the data bytes
                            # later.
                            result_field.data = (
                                self._filename,
                                data_offset,
                                data_depth,
                                data_type,
                            )

                        data_offset += data_depth

                        yield result_field
                except ValueError as valerr:
                    msg = (
                        "Input field skipped as PPField creation failed :"
                        " error = {!r}"
                    )
                    warnings.warn(msg.format(str(valerr)))

    def __iter__(self):
        return pp._interpret_fields(self._extract_field())


def _parse_binary_stream(file_like, dtype=np.float, count=-1):
    """
    Replacement :func:`numpy.fromfile` due to python3 performance issues.

    Args:

    * file_like - Standard python file_like object.

    Kwargs:

    * dtype  - Data type to be parsed out, used to work out bytes read in.

    * count - The number of values required to be generated from the parsing.
        The default is -1, which will read the entire contexts of the file_like
        object and generate as many values as possible.

    """

    # There are a wide range of types supported, we just need to know the byte
    # size of the object, so we just make sure we've go an instance of a
    # np.dtype
    if not isinstance(dtype, np.dtype):
        dtype = np.dtype(dtype)

    # Allocate bytearray for the file to be read into, allowing the numpy array
    # to be writable.
    _buffer = bytearray(count * dtype.itemsize)
    file_like.readinto(_buffer)

    # Let numpy do the heavy lifting once we've sorted the file reading.
    array = np.frombuffer(_buffer, dtype=dtype, count=-1)
    return array


def load_cubes(filenames, callback, constraints=None):
    """
    Loads cubes from a list of fields files filenames.

    Args:

    * filenames - list of fields files filenames to load

    Kwargs:

    * callback - a function which can be passed on to
        :func:`iris.io.run_callback`

    .. note::

        The resultant cubes may not be in the order that they are in the
        file (order is not preserved when there is a field with
        orography references).

    """
    return pp._load_cubes_variable_loader(
        filenames, callback, FF2PP, constraints=constraints
    )


def load_cubes_32bit_ieee(filenames, callback, constraints=None):
    """
    Loads cubes from a list of 32bit ieee converted fieldsfiles filenames.

    .. seealso::

        :func:`load_cubes` for keyword details

    """
    return pp._load_cubes_variable_loader(
        filenames, callback, FF2PP, {"word_depth": 4}, constraints=constraints
    )



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris.analysis.stats.pearsonr` function."""

# Import iris tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import numpy as np
import numpy.ma as ma

import iris
import iris.analysis.stats as stats
from iris.exceptions import CoordinateNotFoundError


@tests.skip_data
class Test(tests.IrisTest):
    def setUp(self):
        # 3D cubes:
        cube_temp = iris.load_cube(
            tests.get_data_path(
                ("NetCDF", "global", "xyt", "SMALL_total_column_co2.nc")
            )
        )
        self.cube_a = cube_temp[0:6]
        self.cube_b = cube_temp[20:26]
        self.cube_b.replace_coord(self.cube_a.coord("time").copy())
        cube_temp = self.cube_a.copy()
        cube_temp.coord("latitude").guess_bounds()
        cube_temp.coord("longitude").guess_bounds()
        self.weights = iris.analysis.cartography.area_weights(cube_temp)

    def test_perfect_corr(self):
        r = stats.pearsonr(self.cube_a, self.cube_a, ["latitude", "longitude"])
        self.assertArrayEqual(r.data, np.array([1.0] * 6))

    def test_perfect_corr_all_dims(self):
        r = stats.pearsonr(self.cube_a, self.cube_a)
        self.assertArrayEqual(r.data, np.array([1.0]))

    def test_incompatible_cubes(self):
        with self.assertRaises(ValueError):
            stats.pearsonr(
                self.cube_a[:, 0, :], self.cube_b[0, :, :], "longitude"
            )

    def test_compatible_cubes(self):
        r = stats.pearsonr(self.cube_a, self.cube_b, ["latitude", "longitude"])
        self.assertArrayAlmostEqual(
            r.data,
            [
                0.81114936,
                0.81690538,
                0.79833135,
                0.81118674,
                0.79745386,
                0.81278484,
            ],
        )

    def test_broadcast_cubes(self):
        r1 = stats.pearsonr(
            self.cube_a, self.cube_b[0, :, :], ["latitude", "longitude"]
        )
        r2 = stats.pearsonr(
            self.cube_b[0, :, :], self.cube_a, ["latitude", "longitude"]
        )
        r_by_slice = [
            stats.pearsonr(
                self.cube_a[i, :, :],
                self.cube_b[0, :, :],
                ["latitude", "longitude"],
            ).data
            for i in range(6)
        ]
        self.assertArrayEqual(r1.data, np.array(r_by_slice))
        self.assertArrayEqual(r2.data, np.array(r_by_slice))

    def test_compatible_cubes_weighted(self):
        r = stats.pearsonr(
            self.cube_a, self.cube_b, ["latitude", "longitude"], self.weights
        )
        self.assertArrayAlmostEqual(
            r.data,
            [
                0.79105429,
                0.79988078,
                0.78825089,
                0.79925653,
                0.79009810,
                0.80115292,
            ],
        )

    def test_broadcast_cubes_weighted(self):
        r = stats.pearsonr(
            self.cube_a,
            self.cube_b[0, :, :],
            ["latitude", "longitude"],
            weights=self.weights[0, :, :],
        )
        r_by_slice = [
            stats.pearsonr(
                self.cube_a[i, :, :],
                self.cube_b[0, :, :],
                ["latitude", "longitude"],
                weights=self.weights[0, :, :],
            ).data
            for i in range(6)
        ]
        self.assertArrayAlmostEqual(r.data, np.array(r_by_slice))

    def test_weight_error(self):
        with self.assertRaises(ValueError):
            stats.pearsonr(
                self.cube_a,
                self.cube_b[0, :, :],
                ["latitude", "longitude"],
                weights=self.weights,
            )

    def test_non_existent_coord(self):
        with self.assertRaises(CoordinateNotFoundError):
            stats.pearsonr(self.cube_a, self.cube_b, "bad_coord")

    def test_mdtol(self):
        cube_small = self.cube_a[:, 0, 0]
        cube_small_masked = cube_small.copy()
        cube_small_masked.data = ma.array(
            cube_small.data, mask=np.array([0, 0, 0, 1, 1, 1], dtype=bool)
        )
        r1 = stats.pearsonr(cube_small, cube_small_masked)
        r2 = stats.pearsonr(cube_small, cube_small_masked, mdtol=0.49)
        self.assertArrayAlmostEqual(r1.data, np.array([0.74586593]))
        self.assertMaskedArrayEqual(r2.data, ma.array([0], mask=[True]))

    def test_common_mask_simple(self):
        cube_small = self.cube_a[:, 0, 0]
        cube_small_masked = cube_small.copy()
        cube_small_masked.data = ma.array(
            cube_small.data, mask=np.array([0, 0, 0, 1, 1, 1], dtype=bool)
        )
        r = stats.pearsonr(cube_small, cube_small_masked, common_mask=True)
        self.assertArrayAlmostEqual(r.data, np.array([1.0]))

    def test_common_mask_broadcast(self):
        cube_small = self.cube_a[:, 0, 0]
        cube_small_2d = self.cube_a[:, 0:2, 0]
        cube_small.data = ma.array(
            cube_small.data, mask=np.array([0, 0, 0, 0, 0, 1], dtype=bool)
        )
        cube_small_2d.data = ma.array(
            np.tile(cube_small.data[:, np.newaxis], 2),
            mask=np.zeros((6, 2), dtype=bool),
        )
        # 2d mask varies on unshared coord:
        cube_small_2d.data.mask[0, 1] = 1
        r = stats.pearsonr(
            cube_small,
            cube_small_2d,
            weights=self.weights[:, 0, 0],
            common_mask=True,
        )
        self.assertArrayAlmostEqual(r.data, np.array([1.0, 1.0]))
        # 2d mask does not vary on unshared coord:
        cube_small_2d.data.mask[0, 0] = 1
        r = stats.pearsonr(cube_small, cube_small_2d, common_mask=True)
        self.assertArrayAlmostEqual(r.data, np.array([1.0, 1.0]))


if __name__ == "__main__":
    tests.main()



import iris
import iris.analysis
import iris.plot as iplt
import matplotlib.pyplot as plt
import matplotlib.colors
import numpy as np

global_air_temp = iris.load_cube(iris.sample_data_path("air_temp.pp"))

regional_ash = iris.load_cube(iris.sample_data_path("NAME_output.txt"))
regional_ash = regional_ash.collapsed("flight_level", iris.analysis.SUM)

# Mask values so low that they are anomalous.
regional_ash.data = np.ma.masked_less(regional_ash.data, 5e-6)

norm = matplotlib.colors.LogNorm(5e-6, 0.0175)

global_air_temp.coord("longitude").guess_bounds()
global_air_temp.coord("latitude").guess_bounds()

fig = plt.figure(figsize=(8, 4.5))

plt.subplot(2, 2, 1)
iplt.pcolormesh(regional_ash, norm=norm)
plt.title("Volcanic ash total\nconcentration not regridded", size="medium")

for subplot_num, mdtol in zip([2, 3, 4], [0, 0.5, 1]):
    plt.subplot(2, 2, subplot_num)
    scheme = iris.analysis.AreaWeighted(mdtol=mdtol)
    global_ash = regional_ash.regrid(global_air_temp, scheme)
    iplt.pcolormesh(global_ash, norm=norm)
    plt.title(
        "Volcanic ash total concentration\n"
        "regridded with AreaWeighted(mdtol={})".format(mdtol),
        size="medium",
    )

plt.subplots_adjust(
    hspace=0, wspace=0.05, left=0.001, right=0.999, bottom=0, top=0.955
)

# Iterate over each of the figure's axes, adding coastlines, gridlines
# and setting the extent.
for ax in fig.axes:
    ax.coastlines("50m")
    ax.gridlines()
    ax.set_extent([-80, 40, 31, 75])

plt.show()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

# Import Iris tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from .gallerytest_util import (
    add_gallery_to_path,
    show_replaced_by_check_graphic,
    fail_any_deprecation_warnings,
)


class TestWindSpeed(tests.GraphicsTest):
    """Test the wind_speed gallery code."""

    def test_plot_wind_speed(self):
        with fail_any_deprecation_warnings():
            with add_gallery_to_path():
                import plot_wind_speed
            with show_replaced_by_check_graphic(self):
                plot_wind_speed.main()


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

import iris.analysis
import iris.tests as tests
import iris.tests.stock
import numpy as np
import numpy.ma as ma


class TestPeakAggregator(tests.IrisTest):
    def test_peak_coord_length_1(self):
        # Coordinate contains a single point.
        latitude = iris.coords.DimCoord(
            np.array([0]), standard_name="latitude", units="degrees"
        )
        cube = iris.cube.Cube(
            np.array([1]), standard_name="air_temperature", units="kelvin"
        )
        cube.add_dim_coord(latitude, 0)

        collapsed_cube = cube.collapsed("latitude", iris.analysis.PEAK)
        self.assertArrayAlmostEqual(
            collapsed_cube.data, np.array([1], dtype=np.float32)
        )

    def test_peak_coord_length_2(self):
        # Coordinate contains 2 points.
        latitude = iris.coords.DimCoord(
            np.arange(0, 2, 1), standard_name="latitude", units="degrees"
        )
        cube = iris.cube.Cube(
            np.array([1, 2]), standard_name="air_temperature", units="kelvin"
        )
        cube.add_dim_coord(latitude, 0)

        collapsed_cube = cube.collapsed("latitude", iris.analysis.PEAK)
        self.assertArrayAlmostEqual(
            collapsed_cube.data, np.array([2], dtype=np.float32)
        )

    def test_peak_coord_length_3(self):
        # Coordinate contains 3 points.
        latitude = iris.coords.DimCoord(
            np.arange(0, 3, 1), standard_name="latitude", units="degrees"
        )
        cube = iris.cube.Cube(
            np.array([1, 2, 1]),
            standard_name="air_temperature",
            units="kelvin",
        )
        cube.add_dim_coord(latitude, 0)

        collapsed_cube = cube.collapsed("latitude", iris.analysis.PEAK)
        self.assertArrayAlmostEqual(
            collapsed_cube.data, np.array([2], dtype=np.float32)
        )

    def test_peak_1d(self):
        # Collapse a 1d cube.
        latitude = iris.coords.DimCoord(
            np.arange(0, 11, 1), standard_name="latitude", units="degrees"
        )
        cube = iris.cube.Cube(
            np.array([1, 2, 3, 4, 5, 6, 5, 4, 3, 2, 1]),
            standard_name="air_temperature",
            units="kelvin",
        )
        cube.add_dim_coord(latitude, 0)

        collapsed_cube = cube.collapsed("latitude", iris.analysis.PEAK)
        self.assertArrayAlmostEqual(
            collapsed_cube.data, np.array([6], dtype=np.float32)
        )

    def test_peak_duplicate_coords(self):
        # Collapse cube along 2 coordinates (both the same).
        latitude = iris.coords.DimCoord(
            np.arange(0, 4, 1), standard_name="latitude", units="degrees"
        )
        cube = iris.cube.Cube(
            np.array([1, 2, 3, 1]),
            standard_name="air_temperature",
            units="kelvin",
        )
        cube.add_dim_coord(latitude, 0)

        collapsed_cube = cube.collapsed("latitude", iris.analysis.PEAK)
        self.assertArrayAlmostEqual(
            collapsed_cube.data, np.array([3], dtype=np.float32)
        )

        collapsed_cube = cube.collapsed(
            ("latitude", "latitude"), iris.analysis.PEAK
        )
        self.assertArrayAlmostEqual(
            collapsed_cube.data, np.array([3], dtype=np.float32)
        )

    def test_peak_2d(self):
        # Collapse a 2d cube.
        longitude = iris.coords.DimCoord(
            np.arange(0, 4, 1), standard_name="longitude", units="degrees"
        )
        latitude = iris.coords.DimCoord(
            np.arange(0, 3, 1), standard_name="latitude", units="degrees"
        )
        cube = iris.cube.Cube(
            np.array([[1, 2, 3, 1], [4, 5, 7, 4], [2, 3, 4, 2]]),
            standard_name="air_temperature",
            units="kelvin",
        )

        cube.add_dim_coord(latitude, 0)
        cube.add_dim_coord(longitude, 1)

        collapsed_cube = cube.collapsed("longitude", iris.analysis.PEAK)
        self.assertArrayAlmostEqual(
            collapsed_cube.data, np.array([3, 7.024054, 4], dtype=np.float32)
        )

        collapsed_cube = cube.collapsed("latitude", iris.analysis.PEAK)
        self.assertArrayAlmostEqual(
            collapsed_cube.data,
            np.array(
                [4.024977, 5.024977, 7.017852, 4.024977], dtype=np.float32
            ),
        )

        collapsed_cube = cube.collapsed(
            ("longitude", "latitude"), iris.analysis.PEAK
        )
        self.assertArrayAlmostEqual(
            collapsed_cube.data, np.array([7.041787], dtype=np.float32)
        )

        collapsed_cube = cube.collapsed(
            ("latitude", "longitude"), iris.analysis.PEAK
        )
        self.assertArrayAlmostEqual(
            collapsed_cube.data, np.array([7.041629], dtype=np.float32)
        )

    def test_peak_without_peak_value(self):
        # No peak in column (values equal).
        latitude = iris.coords.DimCoord(
            np.arange(0, 4, 1), standard_name="latitude", units="degrees"
        )
        cube = iris.cube.Cube(
            np.array([1, 1, 1, 1]),
            standard_name="air_temperature",
            units="kelvin",
        )
        cube.add_dim_coord(latitude, 0)

        collapsed_cube = cube.collapsed("latitude", iris.analysis.PEAK)
        self.assertArrayAlmostEqual(
            collapsed_cube.data, np.array([1], dtype=np.float32)
        )

    def test_peak_with_nan(self):
        # Single nan in column.
        latitude = iris.coords.DimCoord(
            np.arange(0, 5, 1), standard_name="latitude", units="degrees"
        )
        cube = iris.cube.Cube(
            np.array([1, 4, 2, 3, 1], dtype=np.float32),
            standard_name="air_temperature",
            units="kelvin",
        )
        cube.add_dim_coord(latitude, 0)

        cube.data[3] = np.nan

        collapsed_cube = cube.collapsed("latitude", iris.analysis.PEAK)
        self.assertArrayAlmostEqual(
            collapsed_cube.data, np.array([4.024977], dtype=np.float32)
        )
        self.assertEqual(collapsed_cube.data.shape, (1,))

        # Only nans in column.
        cube.data[:] = np.nan

        collapsed_cube = cube.collapsed("latitude", iris.analysis.PEAK)
        self.assertTrue(np.isnan(collapsed_cube.data).all())
        self.assertEqual(collapsed_cube.data.shape, (1,))

    def test_peak_with_mask(self):
        # Single value in column masked.
        latitude = iris.coords.DimCoord(
            np.arange(0, 5, 1), standard_name="latitude", units="degrees"
        )
        cube = iris.cube.Cube(
            ma.array([1, 4, 2, 3, 2], dtype=np.float32),
            standard_name="air_temperature",
            units="kelvin",
        )
        cube.add_dim_coord(latitude, 0)

        cube.data[3] = ma.masked

        collapsed_cube = cube.collapsed("latitude", iris.analysis.PEAK)
        self.assertArrayAlmostEqual(
            collapsed_cube.data, np.array([4.024977], dtype=np.float32)
        )
        self.assertTrue(ma.isMaskedArray(collapsed_cube.data))
        self.assertEqual(collapsed_cube.data.shape, (1,))

        # Whole column masked.
        cube.data[:] = ma.masked

        collapsed_cube = cube.collapsed("latitude", iris.analysis.PEAK)
        masked_array = ma.array(ma.masked)
        self.assertTrue(ma.allequal(collapsed_cube.data, masked_array))
        self.assertTrue(ma.isMaskedArray(collapsed_cube.data))
        self.assertEqual(collapsed_cube.data.shape, (1,))

    def test_peak_with_nan_and_mask(self):
        # Single nan in column with single value masked.
        latitude = iris.coords.DimCoord(
            np.arange(0, 5, 1), standard_name="latitude", units="degrees"
        )
        cube = iris.cube.Cube(
            ma.array([1, 4, 2, 3, 1], dtype=np.float32),
            standard_name="air_temperature",
            units="kelvin",
        )
        cube.add_dim_coord(latitude, 0)

        cube.data[3] = np.nan
        cube.data[4] = ma.masked

        collapsed_cube = cube.collapsed("latitude", iris.analysis.PEAK)
        self.assertArrayAlmostEqual(
            collapsed_cube.data, np.array([4.024977], dtype=np.float32)
        )
        self.assertTrue(ma.isMaskedArray(collapsed_cube.data))
        self.assertEqual(collapsed_cube.data.shape, (1,))

        # Only nans in column where values not masked.
        cube.data[0:3] = np.nan

        collapsed_cube = cube.collapsed("latitude", iris.analysis.PEAK)
        self.assertTrue(np.isnan(collapsed_cube.data).all())
        self.assertTrue(ma.isMaskedArray(collapsed_cube.data))
        self.assertEqual(collapsed_cube.data.shape, (1,))

    def test_peak_against_max(self):
        # Cube with data that infers a peak value greater than the column max.
        latitude = iris.coords.DimCoord(
            np.arange(0, 7, 1), standard_name="latitude", units="degrees"
        )
        cube = iris.cube.Cube(
            np.array([0, 1, 3, 7, 7, 4, 2], dtype=np.float32),
            standard_name="air_temperature",
            units="kelvin",
        )
        cube.add_dim_coord(latitude, 0)

        collapsed_cube = cube.collapsed("latitude", iris.analysis.PEAK)
        self.assertArrayAlmostEqual(
            collapsed_cube.data, np.array([7.630991], dtype=np.float32)
        )

        collapsed_cube = cube.collapsed("latitude", iris.analysis.MAX)
        self.assertArrayAlmostEqual(
            collapsed_cube.data, np.array([7], dtype=np.float32)
        )


if __name__ == "__main__":
    tests.main()
